from __future__ import annotations

from typing import Any

from ..schemas import ToolDefinition
from .base import object_schema
from .file_tools import (
    append_file_handler,
    delete_path_handler,
    list_dir_handler,
    make_dir_handler,
    read_file_handler,
    replace_in_file_handler,
    write_file_handler,
)
from .http_tools import http_get_handler, http_post_json_handler, web_search_handler
from .project_upgrade_tools import build_tools as build_project_upgrade_tools
from .shell_tools import shell_run_handler


def build_builtin_tools(runtime) -> list[ToolDefinition]:
    async def get_status_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.get_session_status(context.session)

    async def set_provider_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.set_provider(context.session, arguments["provider"])

    async def set_model_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.set_model(context.session, arguments["model"])

    async def set_security_mode_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.set_security_mode(context.session, arguments["mode"])

    async def list_providers_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return {"providers": runtime.list_provider_names()}

    async def list_models_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        provider = arguments.get("provider")
        return {"models": await runtime.list_models(provider)}

    async def list_sessions_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return {"sessions": runtime.list_sessions_payload()}

    async def session_end_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.end_session(context.session)

    async def session_reopen_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.reopen_session(context.session)

    async def config_get_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        dotted_path = arguments.get("path")
        if dotted_path:
            return {"path": dotted_path, "value": runtime.config_manager.get(dotted_path)}
        return {"config": runtime.config_manager.data}

    async def config_set_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        runtime.config_manager.set(arguments["path"], arguments.get("value"), backup_reason="config-set-tool")
        runtime.reload()
        return {"path": arguments["path"], "value": runtime.config_manager.get(arguments["path"])}

    async def providers_add_custom_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        payload = {
            "type": arguments.get("type", "openai_compatible"),
            "enabled": bool(arguments.get("enabled", True)),
            "api_key": arguments.get("api_key", ""),
            "base_url": arguments.get("base_url", ""),
            "default_model": arguments.get("default_model", ""),
            "headers": arguments.get("headers", {}) or {},
        }
        runtime.config_manager.set(f"providers.items.{arguments['name']}", payload, backup_reason="add-provider")
        runtime.reload()
        return {"provider": arguments["name"], "config": payload}

    async def reload_runtime_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        runtime.reload()
        return {"reloaded": True, "providers": runtime.list_provider_names()}

    async def reload_skills_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.reload_skills()

    async def list_skills_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return {"skills": runtime.list_skills_payload()}

    async def create_skill_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.skill_manager.create_skill(
            name=arguments["name"],
            description=arguments["description"],
            instructions=arguments.get("instructions"),
            entrypoint_code=arguments.get("entrypoint_code"),
            with_entrypoint=bool(arguments.get("with_entrypoint", True)),
            extra_files=arguments.get("extra_files"),
            context=runtime.skill_context(),
        )

    async def update_skill_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.skill_manager.update_skill(
            name=arguments["name"],
            description=arguments.get("description"),
            instructions=arguments.get("instructions"),
            entrypoint_code=arguments.get("entrypoint_code"),
            manifest_updates=arguments.get("manifest_updates"),
            extra_files=arguments.get("extra_files"),
            context=runtime.skill_context(),
        )

    async def auth_summary_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.auth_summary()

    async def auth_create_pairing_code_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        if not runtime.auth_is_owner(user_id=context.user_id, chat_id=context.chat_id):
            raise PermissionError("Only owners can create pairing codes")
        return runtime.create_pairing_code(
            created_by=context.user_id,
            ttl_seconds=arguments.get("ttl_seconds"),
            note=arguments.get("note", ""),
            max_uses=int(arguments.get("max_uses", 1)),
        )

    async def read_project_agents_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return {"path": str(runtime.project_agents_path()), "content": runtime.read_project_agents_md()}

    async def write_project_agents_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.write_project_agents_md(arguments.get("content", ""))

    async def append_project_agents_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        return runtime.append_project_agents_md(arguments.get("content", ""))

    async def auth_approve_access_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
        if not runtime.auth_is_owner(user_id=context.user_id, chat_id=context.chat_id):
            raise PermissionError("Only owners can approve Telegram access")
        return runtime.approve_telegram_access(
            user_id=arguments.get("user_id"),
            chat_id=arguments.get("chat_id"),
            approved_by=context.user_id,
            username=arguments.get("username"),
            chat_title=arguments.get("chat_title"),
        )

    return [
        ToolDefinition(
            name="shell_run",
            description="Execute a shell command using the current security mode or an explicitly requested mode.",
            parameters=object_schema(
                {
                    "command": {"type": "string", "description": "Shell command to run."},
                    "cwd": {"type": "string", "description": "Optional working directory."},
                    "mode": {"type": "string", "enum": ["safe", "yolo", "god"]},
                    "timeout_seconds": {"type": "integer", "minimum": 0},
                },
                required=["command"],
            ),
            handler=shell_run_handler,
            owner_only=True,
        ),
        ToolDefinition(
            name="read_file",
            description="Read a UTF-8 text file from the project, workspace, data, or logs.",
            parameters=object_schema(
                {
                    "path": {"type": "string"},
                    "max_chars": {"type": "integer", "minimum": 1, "maximum": 200000},
                },
                required=["path"],
            ),
            handler=read_file_handler,
        ),
        ToolDefinition(
            name="write_file",
            description="Write a UTF-8 text file, backing up the previous version if present.",
            parameters=object_schema(
                {
                    "path": {"type": "string"},
                    "content": {"type": "string"},
                    "overwrite": {"type": "boolean"},
                },
                required=["path", "content"],
            ),
            handler=write_file_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="append_file",
            description="Append UTF-8 text to a file, creating it if it does not exist.",
            parameters=object_schema(
                {
                    "path": {"type": "string"},
                    "content": {"type": "string"},
                },
                required=["path", "content"],
            ),
            handler=append_file_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="replace_in_file",
            description="Replace text inside a file, optionally limiting replacement count.",
            parameters=object_schema(
                {
                    "path": {"type": "string"},
                    "old": {"type": "string"},
                    "new": {"type": "string"},
                    "count": {"type": "integer", "minimum": 0},
                },
                required=["path", "old", "new"],
            ),
            handler=replace_in_file_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="list_dir",
            description="List files and directories at a path.",
            parameters=object_schema({"path": {"type": "string"}}),
            handler=list_dir_handler,
        ),
        ToolDefinition(
            name="make_dir",
            description="Create a directory path.",
            parameters=object_schema({"path": {"type": "string"}}, required=["path"]),
            handler=make_dir_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="delete_path",
            description="Delete a file or directory, backing it up first when possible.",
            parameters=object_schema(
                {
                    "path": {"type": "string"},
                    "recursive": {"type": "boolean"},
                },
                required=["path"],
            ),
            handler=delete_path_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="http_get",
            description="Fetch a URL with HTTP GET.",
            parameters=object_schema(
                {
                    "url": {"type": "string"},
                    "headers": {"type": "object"},
                    "timeout_seconds": {"type": "number", "minimum": 1},
                    "max_chars": {"type": "integer", "minimum": 1, "maximum": 200000},
                },
                required=["url"],
            ),
            handler=http_get_handler,
        ),
        ToolDefinition(
            name="http_post_json",
            description="Send a JSON body with HTTP POST.",
            parameters=object_schema(
                {
                    "url": {"type": "string"},
                    "headers": {"type": "object"},
                    "json_body": {"type": "object"},
                    "timeout_seconds": {"type": "number", "minimum": 1},
                    "max_chars": {"type": "integer", "minimum": 1, "maximum": 200000},
                },
                required=["url", "json_body"],
            ),
            handler=http_post_json_handler,
        ),
        ToolDefinition(
            name="web_search",
            description="Search the web using DuckDuckGo HTML and return compact results.",
            parameters=object_schema(
                {
                    "query": {"type": "string"},
                    "max_results": {"type": "integer", "minimum": 1, "maximum": 10},
                },
                required=["query"],
            ),
            handler=web_search_handler,
        ),
        ToolDefinition(
            name="session_get_status",
            description="Get the current session status, model, provider, safety mode, and loaded skills.",
            parameters=object_schema({}),
            handler=get_status_handler,
        ),
        ToolDefinition(
            name="session_list",
            description="List saved sessions known to the runtime.",
            parameters=object_schema({}),
            handler=list_sessions_handler,
        ),
        ToolDefinition(
            name="session_set_provider",
            description="Change the current LLM provider for this session.",
            parameters=object_schema({"provider": {"type": "string"}}, required=["provider"]),
            handler=set_provider_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="session_set_model",
            description="Change the current model for this session.",
            parameters=object_schema({"model": {"type": "string"}}, required=["model"]),
            handler=set_model_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="set_security_mode",
            description="Change the current security mode for shell and file operations. `/safe off` maps to `god`.",
            parameters=object_schema(
                {"mode": {"type": "string", "enum": ["safe", "yolo", "god"]}},
                required=["mode"],
            ),
            handler=set_security_mode_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="session_end",
            description="Close the current session. The agent may call this when it decides the task is done.",
            parameters=object_schema({}),
            handler=session_end_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="session_reopen",
            description="Reopen a previously closed session.",
            parameters=object_schema({}),
            handler=session_reopen_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="read_project_agents_md",
            description="Read the root AGENTS.md file that is injected into the next system prompt if present.",
            parameters=object_schema({}),
            handler=read_project_agents_handler,
        ),
        ToolDefinition(
            name="write_project_agents_md",
            description="Create or replace the root AGENTS.md file. Its contents will be injected into the next system prompt.",
            parameters=object_schema({"content": {"type": "string"}}, required=["content"]),
            handler=write_project_agents_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="append_project_agents_md",
            description="Append instructions to the root AGENTS.md file so they apply on the next generation.",
            parameters=object_schema({"content": {"type": "string"}}, required=["content"]),
            handler=append_project_agents_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="list_providers",
            description="List configured providers.",
            parameters=object_schema({}),
            handler=list_providers_handler,
        ),
        ToolDefinition(
            name="list_models",
            description="List models for one provider or for all providers.",
            parameters=object_schema({"provider": {"type": "string"}}),
            handler=list_models_handler,
        ),
        ToolDefinition(
            name="config_get",
            description="Read a configuration value by dotted path, or return the whole config.",
            parameters=object_schema({"path": {"type": "string"}}),
            handler=config_get_handler,
        ),
        ToolDefinition(
            name="config_set",
            description="Set a configuration value by dotted path and reload the runtime.",
            parameters=object_schema(
                {
                    "path": {"type": "string"},
                    "value": {},
                },
                required=["path", "value"],
            ),
            handler=config_set_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="providers_add_custom",
            description="Add or replace a configured OpenAI-compatible or Gemini provider and reload the runtime.",
            parameters=object_schema(
                {
                    "name": {"type": "string"},
                    "type": {"type": "string", "enum": ["openai_compatible", "custom", "openai", "gemini"]},
                    "api_key": {"type": "string"},
                    "base_url": {"type": "string"},
                    "default_model": {"type": "string"},
                    "enabled": {"type": "boolean"},
                    "headers": {"type": "object"},
                },
                required=["name", "type"],
            ),
            handler=providers_add_custom_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="reload_runtime",
            description="Reload providers, config, and skills without restarting the process.",
            parameters=object_schema({}),
            handler=reload_runtime_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="reload_skills",
            description="Hot reload all workspace skills immediately.",
            parameters=object_schema({}),
            handler=reload_skills_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="list_skills",
            description="List all known skills and their status.",
            parameters=object_schema({}),
            handler=list_skills_handler,
        ),
        ToolDefinition(
            name="create_skill",
            description="Create a new workspace skill with SKILL.md, manifest, and optional Python entrypoint.",
            parameters=object_schema(
                {
                    "name": {"type": "string"},
                    "description": {"type": "string"},
                    "instructions": {"type": "string"},
                    "entrypoint_code": {"type": "string"},
                    "with_entrypoint": {"type": "boolean"},
                    "extra_files": {"type": "object", "additionalProperties": {"type": "string"}},
                },
                required=["name", "description"],
            ),
            handler=create_skill_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="update_skill",
            description="Update an existing skill and hot reload it.",
            parameters=object_schema(
                {
                    "name": {"type": "string"},
                    "description": {"type": "string"},
                    "instructions": {"type": "string"},
                    "entrypoint_code": {"type": "string"},
                    "manifest_updates": {"type": "object"},
                    "extra_files": {"type": "object", "additionalProperties": {"type": "string"}},
                },
                required=["name"],
            ),
            handler=update_skill_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="auth_summary",
            description="Show Telegram access state: owners, granted users/chats, and pending pairing codes.",
            parameters=object_schema({}),
            handler=auth_summary_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="auth_create_pairing_code",
            description="Create a Telegram pairing code for granting access to the bot.",
            parameters=object_schema(
                {
                    "ttl_seconds": {"type": "integer", "minimum": 1},
                    "max_uses": {"type": "integer", "minimum": 1},
                    "note": {"type": "string"},
                }
            ),
            handler=auth_create_pairing_code_handler,
            destructive=True,
        ),
        ToolDefinition(
            name="auth_approve_access",
            description="Approve a Telegram user/chat manually.",
            parameters=object_schema(
                {
                    "user_id": {"type": "integer"},
                    "chat_id": {"type": "integer"},
                    "username": {"type": "string"},
                    "chat_title": {"type": "string"},
                }
            ),
            handler=auth_approve_access_handler,
            destructive=True,
        ),
    ] + build_project_upgrade_tools()



from __future__ import annotations

from pathlib import Path
from typing import Any

from .builder import create_skill as build_skill
from .builder import update_skill as patch_skill
from .runtime import SkillRegistry


class SkillView(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as exc:
            raise AttributeError(item) from exc



def _resolve_skills_root(config: dict[str, Any]) -> Path:
    if config.get("skills_root"):
        return Path(config["skills_root"]).expanduser().resolve()

    workspace = config.get("workspace", {}) or {}
    if workspace.get("skills_root"):
        return Path(workspace["skills_root"]).expanduser().resolve()
    if workspace.get("skills_dir"):
        return Path(workspace["skills_dir"]).expanduser().resolve()
    if workspace.get("root"):
        return (Path(workspace["root"]).expanduser().resolve() / "skills").resolve()

    return Path("workspace/skills").expanduser().resolve()


class SkillManager:
    def __init__(self, config: dict[str, Any]):
        self.config = dict(config)
        self.root = _resolve_skills_root(self.config)
        self.registry = SkillRegistry(self.root)

    def _serialize_skill(self, name: str) -> dict[str, Any]:
        skill = self.registry.get(name)
        if skill is None:
            return SkillView({"name": name, "error": self.registry.errors.get(name)})
        return SkillView(
            {
                "name": skill.manifest.name,
                "description": skill.manifest.description,
                "version": skill.manifest.version,
                "enabled": skill.manifest.enabled,
                "fingerprint": skill.fingerprint,
                "path": str(skill.path),
                "tools": skill.export.tools,
                "prompts": skill.export.prompts,
                "metadata": skill.export.metadata,
                "bundle_files": skill.manifest.metadata.get("bundle_files", []),
                "error": self.registry.errors.get(name),
            }
        )

    def load_all(self, *, context: dict[str, Any] | None = None) -> list[dict[str, Any]]:
        self.registry = SkillRegistry(self.root)
        self.registry.reload(context=context)
        return [self._serialize_skill(name) for name in sorted(self.registry.skills)]

    def reload_if_needed(self, *, context: dict[str, Any] | None = None):
        return self.registry.reload(context=context)

    def list_skills(self) -> list[dict[str, Any]]:
        names = sorted(set(self.registry.skills) | set(self.registry.errors))
        return [self._serialize_skill(name) for name in names]

    def get_system_prompt_fragments(self) -> list[str]:
        return self.registry.collect_prompt_fragments()

    def get_tools(self) -> list[dict[str, Any]]:
        return self.registry.collect_tools()

    def create_skill(
        self,
        *,
        name: str,
        description: str,
        instructions: str | None = None,
        entrypoint_code: str | None = None,
        with_entrypoint: bool = True,
        extra_files: dict[str, str] | None = None,
        context: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        skill_dir = build_skill(
            self.root,
            name=name,
            description=description,
            instructions=instructions,
            entrypoint_code=entrypoint_code,
            with_entrypoint=with_entrypoint,
            extra_files=extra_files,
        )
        summary = self.reload_if_needed(context=context)
        return {
            "skill_dir": str(skill_dir),
            "extra_files": sorted((extra_files or {}).keys()),
            "reload": {
                "loaded": summary.loaded,
                "reloaded": summary.reloaded,
                "removed": summary.removed,
                "failed": summary.failed,
            },
        }

    def update_skill(
        self,
        *,
        name: str,
        description: str | None = None,
        instructions: str | None = None,
        entrypoint_code: str | None = None,
        manifest_updates: dict[str, Any] | None = None,
        extra_files: dict[str, str] | None = None,
        context: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        result = patch_skill(
            self.root / name,
            description=description,
            instructions=instructions,
            entrypoint_code=entrypoint_code,
            manifest_updates=manifest_updates,
            extra_files=extra_files,
        )
        summary = self.reload_if_needed(context=context)
        return {
            "skill_dir": str(result["skill_dir"]),
            "backup_dir": str(result["backup_dir"]) if result["backup_dir"] else None,
            "extra_files": sorted((extra_files or {}).keys()),
            "reload": {
                "loaded": summary.loaded,
                "reloaded": summary.reloaded,
                "removed": summary.removed,
                "failed": summary.failed,
            },
        }



from __future__ import annotations

from types import SimpleNamespace

from openclaw_lite.interfaces.telegram_media import (
    TelegramInboundMedia,
    build_inbox_root,
    build_media_save_path,
    extract_telegram_media,
    format_media_prompt_payload,
    make_fake_message,
)


def test_build_media_save_path_uses_chat_message_and_kind(tmp_path) -> None:
    item = TelegramInboundMedia(kind="document", file_id="f1", file_unique_id="u1", original_name="Report 2026.txt")
    path = build_media_save_path(tmp_path, chat_id=-100123, message_id=77, item=item)
    assert path.parent.name == "-100123"
    assert path.name == "msg77-document-Report-2026.txt"


def test_extract_telegram_media_reads_text_and_multiple_kinds() -> None:
    message = make_fake_message(
        text="process this",
        caption=None,
        photo=[SimpleNamespace(file_id="p1", file_unique_id="pu1", file_size=123)],
        document=SimpleNamespace(file_id="d1", file_unique_id="du1", file_name="hello.pdf", mime_type="application/pdf", file_size=456),
        video=None,
        animation=None,
        audio=None,
        voice=None,
        sticker=None,
    )
    items = extract_telegram_media(message)
    assert [item.kind for item in items] == ["photo", "document"]
    assert items[0].message_text == "process this"
    assert items[1].original_name == "hello.pdf"


def test_format_media_prompt_payload_contains_text_and_paths() -> None:
    payload = format_media_prompt_payload(
        message_text="convert this file",
        saved_items=[
            {
                "kind": "document",
                "saved_path": "/data/telegram_inbox/1/msg7-document-file.pdf",
                "original_name": "file.pdf",
            }
        ],
    )
    assert "convert this file" in payload
    assert "[saved telegram attachments]" in payload
    assert "path=/data/telegram_inbox/1/msg7-document-file.pdf" in payload


def test_build_inbox_root_uses_data_dir(tmp_path) -> None:
    inbox = build_inbox_root({"data_dir": str(tmp_path)})
    assert inbox == tmp_path / "telegram_inbox"



from __future__ import annotations

from .base import (
    BaseProvider,
    JsonDict,
    ProviderError,
    build_gemini_contents,
    build_gemini_tools,
    ensure_trailing_slashless,
    extract_gemini_model_names,
    gemini_system_instruction,
    normalise_gemini_response,
    with_query_params,
)


class GeminiProvider(BaseProvider):
    provider_type = "gemini"

    def __init__(
        self,
        *,
        api_key: str,
        base_url: str = "https://generativelanguage.googleapis.com/v1beta",
        provider_name: str = "gemini",
        timeout: float = 60.0,
        http_client=None,
    ) -> None:
        super().__init__(provider_name=provider_name, timeout=timeout, http_client=http_client)
        self.api_key = api_key
        self.base_url = ensure_trailing_slashless(base_url)

    def _ensure_configured(self) -> None:
        if not self.api_key:
            raise ProviderError("gemini requires an API key. Configure it via Telegram /settings or /apikey gemini <key>.")

    async def generate(
        self,
        messages: list[JsonDict],
        tools: list[JsonDict] | None,
        model: str,
        system_prompt: str | None = None,
        temperature: float | None = None,
    ) -> JsonDict:
        self._ensure_configured()
        payload: JsonDict = {"contents": build_gemini_contents(messages)}
        tool_block = build_gemini_tools(tools)
        if tool_block:
            payload["tools"] = tool_block
        system_instruction = gemini_system_instruction(system_prompt)
        if system_instruction:
            payload["systemInstruction"] = system_instruction
        if temperature is not None:
            payload["generationConfig"] = {"temperature": temperature}

        url = with_query_params(f"{self.base_url}/models/{model}:generateContent", key=self.api_key)
        response = await self.http.post_json(url, payload=payload)
        return normalise_gemini_response(response.payload, fallback_model=model, provider_name=self.provider_name)

    async def list_models(self) -> list[str]:
        self._ensure_configured()
        url = with_query_params(f"{self.base_url}/models", key=self.api_key)
        response = await self.http.get_json(url)
        return extract_gemini_model_names(response.payload)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

from ..auth import AuthManager
from ..config import ConfigManager, init_project_layout
from ..providers import build_provider_registry
from ..schemas import SessionState, TELEGRAM_ACTIVE_USER_ID_METADATA_KEY, ToolDefinition
from ..skills.manager import SkillManager
from ..tools.builtin import build_builtin_tools
from .agent import AgentLoop
from .session_store import SessionStore
from .shell import ShellExecutor
from .tool_registry import ToolRegistry


class OpenClawLiteRuntime:
    def __init__(self, config_path: str | Path = "config.yaml"):
        self.config_manager = ConfigManager(config_path)
        init_project_layout(self.config_manager)
        self.config = self.config_manager.data
        self.auth_manager = AuthManager(self.config)
        self.shell_executor = ShellExecutor(self.config)
        self.provider_registry = build_provider_registry(self.config)
        self.skill_manager = SkillManager(self.skill_config())
        self.skill_manager.load_all(context=self.skill_context())
        self.session_store = SessionStore(
            self.config.get("data_dir", "data"),
            default_provider=self.default_provider_name(),
            default_model=self.default_model_for_provider(self.default_provider_name()),
            history_limit=int(self.config.get("history_limit", 0)),
        )
        self.tool_registry = ToolRegistry(Path(self.config.get("logs_dir", "logs")) / "tool_audit.log")
        self._registered_dynamic_skill_tools: set[str] = set()
        self.external_tool_registrars: list[Callable[[ToolRegistry], None]] = []
        self.telegram_bot = None
        self.source_root = Path(__file__).resolve().parents[2]
        self.config["source_root"] = str(self.source_root)
        self._register_tools()
        self.agent_loop = AgentLoop(self)

    def skill_config(self) -> dict[str, Any]:
        workspace_root = Path(self.config.get("workspace_dir", "workspace")).resolve()
        skills_root = workspace_root / "skills"
        return {
            "skills_root": str(skills_root),
            "workspace": {
                "root": str(workspace_root),
                "skills_root": str(skills_root),
            },
        }

    def skill_context(self) -> dict[str, Any]:
        return {
            "runtime": self,
            "project_root": self.config.get("project_root"),
            "workspace_dir": self.config.get("workspace_dir"),
            "logs_dir": self.config.get("logs_dir"),
        }

    def add_tool_registrar(self, registrar: Callable[[ToolRegistry], None]) -> None:
        self.external_tool_registrars.append(registrar)
        registrar(self.tool_registry)

    def _register_tools(self) -> None:
        self.tool_registry = ToolRegistry(Path(self.config.get("logs_dir", "logs")) / "tool_audit.log")
        self.tool_registry.register_many(build_builtin_tools(self))
        self.sync_skill_tools()
        for registrar in self.external_tool_registrars:
            registrar(self.tool_registry)

    def sync_skill_tools(self) -> None:
        self._registered_dynamic_skill_tools = set()
        for item in self.skill_manager.get_tools():
            if not isinstance(item, dict):
                continue
            name = item.get("name")
            handler = item.get("handler")
            if not name or not callable(handler):
                continue
            definition = ToolDefinition(
                name=name,
                description=item.get("description", f"Skill tool: {name}"),
                parameters=item.get("parameters", {"type": "object", "properties": {}}),
                handler=handler,
                owner_only=bool(item.get("owner_only", True)),
                destructive=bool(item.get("destructive", False)),
            )
            self.tool_registry.register(definition)
            self._registered_dynamic_skill_tools.add(name)

    def reload(self) -> None:
        self.config = self.config_manager.reload()
        init_project_layout(self.config_manager)
        self.auth_manager = AuthManager(self.config)
        self.shell_executor = ShellExecutor(self.config)
        self.provider_registry = build_provider_registry(self.config)
        self.skill_manager = SkillManager(self.skill_config())
        self.skill_manager.load_all(context=self.skill_context())
        self.session_store.default_provider = self.default_provider_name()
        self.session_store.default_model = self.default_model_for_provider(self.default_provider_name())
        self._register_tools()

    def reload_skills(self) -> dict[str, Any]:
        summary = self.skill_manager.reload_if_needed(context=self.skill_context())
        self._register_tools()
        return {
            "loaded": summary.loaded,
            "reloaded": summary.reloaded,
            "removed": summary.removed,
            "failed": summary.failed,
        }

    def is_owner(self, user_id: int | None, chat_id: int | str | None = None) -> bool:
        resolved_chat_id = int(chat_id) if chat_id is not None and str(chat_id).lstrip("-").isdigit() else None
        return self.auth_manager.is_owner(user_id, resolved_chat_id)

    def is_authorized(self, user_id: int | None, chat_id: int | str | None = None) -> bool:
        resolved_chat_id = int(chat_id) if chat_id is not None and str(chat_id).lstrip("-").isdigit() else None
        return self.auth_manager.is_allowed(user_id, resolved_chat_id)

    def generate_pairing_code(self, *, created_by: int | None = None, note: str | None = None) -> dict[str, Any]:
        return self.auth_manager.generate_pairing_code(created_by=created_by, note=note)

    def claim_pairing_code(self, code: str, *, user_id: int | None, chat_id: int | str | None) -> dict[str, Any]:
        resolved_chat_id = int(chat_id) if chat_id is not None and str(chat_id).lstrip("-").isdigit() else None
        return self.auth_manager.claim_pairing_code(code, user_id=user_id, chat_id=resolved_chat_id)

    def auth_summary(self) -> dict[str, Any]:
        summary = self.auth_manager.summary()
        allowed_users = summary.get("allowed_user_ids", [])
        allowed_chats = summary.get("allowed_chat_ids", [])
        pairing_codes = summary.get("pairing_codes", summary.get("pending_codes", []))
        return {
            "owner_user_ids": summary.get("owner_user_ids", []),
            "owner_chat_ids": summary.get("owner_chat_ids", []),
            "allowed_user_ids": allowed_users,
            "allowed_chat_ids": allowed_chats,
            "approved_users": allowed_users,
            "approved_chats": allowed_chats,
            "pending_codes": pairing_codes,
            "pairing_codes": pairing_codes,
            "grants": summary.get("grants", []),
            "state_path": summary.get("state_path"),
            "bootstrap_code_configured": summary.get("bootstrap_code_configured", False),
        }

    def create_pairing_code(
        self,
        *,
        created_by_user_id: int | None = None,
        created_by: int | None = None,
        ttl_sec: int | None = None,
        ttl_seconds: int | None = None,
        note: str | None = None,
        max_uses: int | None = None,
    ) -> dict[str, Any]:
        creator = created_by_user_id if created_by_user_id is not None else created_by
        ttl = ttl_sec if ttl_sec is not None else ttl_seconds
        create_method = getattr(self.auth_manager, "create_pairing_code", None)
        if callable(create_method):
            result = create_method(created_by=creator, expires_in=ttl, note=note or "", max_uses=max_uses or 1)
            if hasattr(result, "__dict__"):
                return dict(result.__dict__)
        return self.auth_manager.generate_pairing_code(created_by=creator, ttl_seconds=ttl, note=note)

    def redeem_pairing_code(self, code: str, *, user_id: int | None, chat_id: int | str | None) -> dict[str, Any]:
        result = self.claim_pairing_code(code, user_id=user_id, chat_id=chat_id)
        return {"authorized": True, **result}

    def user_is_admin(self, *, user_id: int | None, chat_id: int | str | None = None) -> bool:
        return self.is_authorized(user_id, chat_id)

    def user_has_access(self, *, user_id: int | None, chat_id: int | str | None = None) -> bool:
        return self.is_authorized(user_id, chat_id)

    def auth_is_owner(self, *, user_id: int | None = None, chat_id: int | str | None = None) -> bool:
        return self.is_owner(user_id, chat_id)

    def auth_is_allowed(self, *, user_id: int | None = None, chat_id: int | str | None = None) -> bool:
        return self.is_authorized(user_id, chat_id)

    def pair_telegram_access(self, code: str, *, user_id: int | None, chat_id: int | str | None, username: str | None = None, chat_title: str | None = None) -> dict[str, Any]:
        payload = self.claim_pairing_code(code, user_id=user_id, chat_id=chat_id)
        if username is not None:
            payload["username"] = username
        if chat_title is not None:
            payload["chat_title"] = chat_title
        return payload

    def approve_telegram_access(self, *, user_id: int | None = None, chat_id: int | str | None = None, approved_by: int | None = None, username: str | None = None, chat_title: str | None = None) -> dict[str, Any]:
        resolved_chat_id = int(chat_id) if chat_id is not None and str(chat_id).lstrip("-").isdigit() else None
        return self.auth_manager.approve_identity(
            user_id=user_id,
            chat_id=resolved_chat_id,
            via="telegram_owner_approve",
            paired_by=approved_by,
            username=username,
            chat_title=chat_title,
        )

    def authorize_user(self, user_id: int, *, chat_id: int | None = None, source: str = "manual") -> dict[str, Any]:
        return self.auth_manager.approve_identity(user_id=user_id, chat_id=chat_id, via=source)

    def revoke_user(self, user_id: int) -> dict[str, Any]:
        return self.auth_manager.revoke(user_id=user_id)

    def project_agents_path(self) -> Path:
        return Path(str(self.config.get("project_root", "."))).expanduser().resolve() / "AGENTS.md"

    def read_project_agents_md(self) -> str:
        path = self.project_agents_path()
        if not path.exists():
            return ""
        return path.read_text(encoding="utf-8")

    def write_project_agents_md(self, content: str) -> dict[str, Any]:
        path = self.project_agents_path()
        path.write_text(content, encoding="utf-8")
        return {"path": str(path), "bytes": path.stat().st_size}

    def append_project_agents_md(self, content: str) -> dict[str, Any]:
        path = self.project_agents_path()
        existing = path.read_text(encoding="utf-8") if path.exists() else ""
        updated = (existing.rstrip() + "\n\n" + content.strip() + "\n").lstrip()
        path.write_text(updated, encoding="utf-8")
        return {"path": str(path), "bytes": path.stat().st_size}

    def root_agents_path(self) -> Path:
        return self.project_agents_path()

    def read_root_agents_md(self) -> str:
        return self.read_project_agents_md()

    def write_root_agents_md(self, content: str) -> dict[str, Any]:
        return self.write_project_agents_md(content)

    def append_root_agents_md(self, content: str) -> dict[str, Any]:
        return self.append_project_agents_md(content)

    def list_access_payload(self) -> dict[str, Any]:
        summary = self.auth_summary()
        return {
            "owner_user_ids": summary.get("owner_user_ids", []),
            "owner_chat_ids": summary.get("owner_chat_ids", []),
            "authorized_user_ids": summary.get("approved_users", []),
            "authorized_chat_ids": summary.get("approved_chats", []),
            "pending_codes": summary.get("pending_codes", []),
            "grants": summary.get("grants", []),
            "bootstrap_code_configured": summary.get("bootstrap_code_configured", False),
            "state_path": summary.get("state_path"),
        }

    def default_provider_name(self) -> str:
        return str(self.config.get("providers", {}).get("default", self.config_manager.default_provider()))

    def default_model_for_provider(self, provider_name: str) -> str:
        provider = (self.config.get("providers", {}).get("items", {}) or {}).get(provider_name, {})
        return str(provider.get("default_model") or self.config_manager.default_model())

    def provider_catalog(self) -> dict[str, dict[str, Any]]:
        items = (self.config.get("providers", {}) or {}).get("items", {}) or {}
        return {str(name): dict(value) for name, value in items.items() if isinstance(value, dict)}

    def provider_catalog_names(self) -> list[str]:
        return sorted(self.provider_catalog())

    def provider_summary_payload(self) -> list[dict[str, Any]]:
        default_name = str((self.config.get("providers", {}) or {}).get("default", self.default_provider_name()))
        payload = []
        for name, item in self.provider_catalog().items():
            payload.append(
                {
                    "name": name,
                    "type": item.get("type"),
                    "enabled": item.get("enabled", True),
                    "has_api_key": bool(item.get("api_key")),
                    "default_model": item.get("default_model"),
                    "is_default": name == default_name,
                }
            )
        return payload

    def configure_provider(
        self,
        provider_name: str,
        *,
        api_key: str | None = None,
        default_model: str | None = None,
        enabled: bool | None = None,
        make_default: bool | None = None,
    ) -> dict[str, Any]:
        items = self.provider_catalog()
        entry = dict(items.get(provider_name, {}))
        if not entry:
            entry = {"type": provider_name if provider_name in {"gemini", "claude", "anthropic", "openai"} else "openai_compatible"}
        if api_key is not None:
            entry["api_key"] = api_key
        if default_model is not None:
            entry["default_model"] = default_model
        if enabled is not None:
            entry["enabled"] = enabled
        elif api_key is not None:
            entry["enabled"] = True
        self.config_manager.set(f"providers.items.{provider_name}", entry, backup_reason="provider-config")
        if make_default:
            self.config_manager.set("providers.default", provider_name, backup_reason="provider-default")
        self.reload()
        return {"provider": provider_name, "config": self.provider_catalog().get(provider_name, {})}

    def set_provider_api_key(self, provider_name: str, api_key: str) -> dict[str, Any]:
        return self.configure_provider(provider_name, api_key=api_key, enabled=True)

    def set_provider_default_model(self, provider_name: str, model: str) -> dict[str, Any]:
        return self.configure_provider(provider_name, default_model=model)

    def set_default_provider_config(self, provider_name: str) -> dict[str, Any]:
        self.config_manager.set("providers.default", provider_name, backup_reason="provider-default")
        self.reload()
        return {"default_provider": provider_name}

    def list_provider_names(self) -> list[str]:
        return self.provider_registry.list_provider_names()

    async def list_models(self, provider_name: str | None = None) -> dict[str, Any] | list[str]:
        if provider_name:
            try:
                return await self.provider_registry.get(provider_name).list_models()
            except Exception:
                entry = self.provider_catalog().get(provider_name, {})
                fallback = []
                if entry.get("default_model"):
                    fallback.append(entry.get("default_model"))
                if isinstance(entry.get("models"), list):
                    fallback.extend(str(item) for item in entry.get("models") if item)
                return sorted(dict.fromkeys(fallback))
        payload = {}
        for name in self.provider_catalog_names():
            payload[name] = await self.list_models(name)
        return payload

    def get_session(self, *, session_id: str, chat_id: str, user_id: int | None) -> SessionState:
        session = self.session_store.get_or_create(session_id, chat_id=chat_id, user_id=user_id)
        if not session.provider:
            session.provider = self.default_provider_name()
        if not session.model:
            session.model = self.default_model_for_provider(session.provider)
        self.session_store.save_state(session)
        return session

    def serialise_skill_status(self, status: Any) -> dict[str, Any]:
        if isinstance(status, dict):
            return status
        return {
            "name": getattr(status, "name", str(status)),
            "enabled": getattr(status, "enabled", False),
            "fingerprint": getattr(status, "fingerprint", None),
            "error": getattr(status, "error", None),
        }

    def list_skills_payload(self) -> list[dict[str, Any]]:
        return [self.serialise_skill_status(item) for item in self.skill_manager.list_skills()]

    def get_session_status(self, session: SessionState) -> dict[str, Any]:
        return {
            "session_id": session.session_id,
            "chat_id": session.chat_id,
            "provider": session.provider,
            "model": session.model,
            "security_mode": session.security_mode,
            "history_messages": len(session.history),
            "closed": session.closed,
            "skills": self.list_skills_payload(),
            "project_root": self.config.get("project_root"),
            "workspace_dir": self.config.get("workspace_dir"),
        }

    def set_security_mode(self, session: SessionState, mode: str) -> dict[str, Any]:
        if mode not in {"safe", "yolo", "god"}:
            raise ValueError(f"Unsupported security mode: {mode}")
        session.security_mode = mode  # type: ignore[assignment]
        self.session_store.save_state(session)
        return {"security_mode": session.security_mode, "unrestricted": session.security_mode == "god"}

    def set_provider(self, session: SessionState, provider_name: str) -> dict[str, Any]:
        if provider_name not in self.list_provider_names():
            raise ValueError(f"Unknown provider: {provider_name}")
        session.provider = provider_name
        session.model = self.default_model_for_provider(provider_name)
        self.session_store.save_state(session)
        return {"provider": session.provider, "model": session.model}

    def set_model(self, session: SessionState, model: str) -> dict[str, Any]:
        session.model = model
        self.session_store.save_state(session)
        return {"provider": session.provider, "model": session.model}

    def max_agent_steps(self) -> int | None:
        raw = self.config.get("tool_loop_limit", 0)
        if raw in (None, 0, "0", False):
            return None
        return int(raw)

    def list_sessions_payload(self) -> list[dict[str, Any]]:
        return [self.get_session_status(session) for session in self.session_store.list_sessions()]

    def end_session(self, session: SessionState) -> dict[str, Any]:
        session.closed = True
        session.metadata.pop(TELEGRAM_ACTIVE_USER_ID_METADATA_KEY, None)
        self.session_store.save_state(session)
        return {"session_id": session.session_id, "closed": True}

    def reopen_session(self, session: SessionState) -> dict[str, Any]:
        session.closed = False
        self.session_store.save_state(session)
        return {"session_id": session.session_id, "closed": False}

    def reset_session(self, session_id: str) -> dict[str, Any]:
        self.session_store.reset(session_id)
        return {"session_id": session_id, "reset": True}

    async def handle_user_message(
        self,
        *,
        session_id: str,
        chat_id: str,
        user_id: int | None,
        text: str,
        transport: str = "local",
        transport_state: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        return await self.agent_loop.run_turn(
            session_id=session_id,
            chat_id=chat_id,
            user_id=user_id,
            text=text,
            transport=transport,
            transport_state=transport_state,
        )

    async def run(self) -> None:
        if not self.config.get("telegram", {}).get("enabled", True):
            raise RuntimeError("Telegram is disabled in config")
        from ..interfaces.telegram_bot import TelegramBotInterface

        interface = TelegramBotInterface(self)
        await interface.run()

    def doctor(self) -> dict[str, Any]:
        providers = {}
        for name, raw in (self.config.get("providers", {}).get("items", {}) or {}).items():
            providers[name] = {
                "type": raw.get("type"),
                "enabled": raw.get("enabled", True),
                "has_api_key": bool(raw.get("api_key")),
                "default_model": raw.get("default_model"),
            }
        return {
            "project_root": self.config.get("project_root"),
            "workspace_dir": self.config.get("workspace_dir"),
            "skills_root": self.skill_config()["skills_root"],
            "telegram_enabled": self.config.get("telegram", {}).get("enabled", True),
            "telegram_has_token": bool(self.config.get("telegram", {}).get("bot_token")),
            "provider_names": self.list_provider_names(),
            "providers": providers,
            "skills": self.list_skills_payload(),
            "auth": self.auth_summary(),
            "access": self.list_access_payload(),
            "tool_loop_limit": self.max_agent_steps(),
            "history_limit": self.session_store.history_limit,
            "shell_timeout_seconds": self.shell_executor.timeout_seconds,
            "shell_output_limit": self.shell_executor.max_output_chars,
            "liteclaw_command": "liteclaw",
        }



from __future__ import annotations

import asyncio
import os
import shlex
from dataclasses import dataclass
from pathlib import Path

from openclaw_lite.config import AppConfig
from openclaw_lite.schema import ToolContext, ToolResult, ToolSpec
from openclaw_lite.tools.base import fail, ok
from openclaw_lite.utils import shorten


@dataclass(slots=True)
class ShellExecution:
    command: str
    returncode: int
    stdout: str
    stderr: str
    mode: str
    cwd: str


def _is_safe_command(command: str, config: AppConfig) -> tuple[bool, str | None]:
    lowered = command.strip().lower()
    for forbidden in config.shell.safe_blocklist:
        if forbidden in lowered:
            return False, f"blocked by pattern: {forbidden}"
    try:
        argv = shlex.split(command)
    except ValueError as exc:
        return False, f"shlex parse error: {exc}"
    if not argv:
        return False, "empty command"
    executable = Path(argv[0]).name
    if executable not in config.shell.safe_allowlist:
        return False, f"command `{executable}` not in safe allowlist"
    return True, None


async def run_shell_command(
    config: AppConfig,
    command: str,
    mode: str,
    cwd: str | Path | None = None,
) -> ShellExecution:
    target_cwd = Path(cwd or config.paths.workspace).resolve()
    target_cwd.mkdir(parents=True, exist_ok=True)
    env = os.environ.copy()

    if mode == "safe":
        allowed, reason = _is_safe_command(command, config)
        if not allowed:
            return ShellExecution(command=command, returncode=126, stdout="", stderr=reason or "unsafe", mode=mode, cwd=str(target_cwd))
        argv = shlex.split(command)
        proc = await asyncio.create_subprocess_exec(
            *argv,
            cwd=str(target_cwd),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env,
        )
    else:
        proc = await asyncio.create_subprocess_shell(
            command,
            cwd=str(target_cwd),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env,
        )

    try:
        stdout_b, stderr_b = await asyncio.wait_for(proc.communicate(), timeout=config.shell.timeout_sec)
    except asyncio.TimeoutError:
        proc.kill()
        await proc.wait()
        return ShellExecution(
            command=command,
            returncode=124,
            stdout="",
            stderr=f"timeout after {config.shell.timeout_sec}s",
            mode=mode,
            cwd=str(target_cwd),
        )

    stdout = shorten(stdout_b.decode("utf-8", errors="replace"), config.shell.max_output_chars)
    stderr = shorten(stderr_b.decode("utf-8", errors="replace"), config.shell.max_output_chars)
    return ShellExecution(
        command=command,
        returncode=proc.returncode,
        stdout=stdout,
        stderr=stderr,
        mode=mode,
        cwd=str(target_cwd),
    )


async def shell_tool(context: ToolContext, arguments: dict) -> ToolResult:
    command = str(arguments.get("command", "")).strip()
    cwd = arguments.get("cwd") or context.config.paths.workspace
    result = await run_shell_command(context.config, command=command, mode=context.session.mode, cwd=cwd)
    return ok(
        {
            "command": result.command,
            "returncode": result.returncode,
            "stdout": result.stdout,
            "stderr": result.stderr,
            "mode": result.mode,
            "cwd": result.cwd,
        }
    )


def build_tools() -> list[ToolSpec]:
    return [
        ToolSpec(
            name="shell_exec",
            description="Run a shell command in the current workspace. Mode depends on the active session security mode.",
            parameters={
                "type": "object",
                "properties": {
                    "command": {"type": "string", "description": "Shell command to execute."},
                    "cwd": {"type": "string", "description": "Optional working directory."},
                },
                "required": ["command"],
            },
            handler=shell_tool,
            dangerous=True,
        )
    ]



from __future__ import annotations

import asyncio
import html
import json
import re
from typing import Any
from urllib import error, parse, request
from urllib.parse import parse_qs, quote_plus, unquote, urlparse

from ..util import truncate_text

TAG_RE = re.compile(r"<[^>]+>")
RESULT_RE = re.compile(
    r'<a[^>]*class="result__a"[^>]*href="(?P<href>[^"]+)"[^>]*>(?P<title>.*?)</a>.*?(?:<a[^>]*class="result__snippet"[^>]*>(?P<snippet_a>.*?)</a>|<div[^>]*class="result__snippet"[^>]*>(?P<snippet_b>.*?)</div>)',
    re.S,
)


def _strip_tags(value: str) -> str:
    return html.unescape(TAG_RE.sub("", value)).strip()


def _normalize_duckduckgo_url(url: str) -> str:
    parsed = urlparse(url)
    if parsed.netloc.endswith("duckduckgo.com") and parsed.path.startswith("/l/"):
        query = parse_qs(parsed.query)
        if "uddg" in query:
            return unquote(query["uddg"][0])
    return url


def _http_request(url: str, *, method: str, headers: dict[str, str] | None = None, json_body: dict[str, Any] | None = None) -> tuple[int, dict[str, str], str]:
    payload = None
    merged_headers = dict(headers or {})
    if json_body is not None:
        merged_headers.setdefault("Content-Type", "application/json")
        payload = json.dumps(json_body, ensure_ascii=False).encode("utf-8")
    req = request.Request(url=url, method=method, headers=merged_headers, data=payload)
    with request.urlopen(req, timeout=30) as response:
        body = response.read().decode("utf-8", errors="replace")
        return response.status, dict(response.headers.items()), body


async def http_get_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
    max_chars = int(arguments.get("max_chars", 12000))
    status, headers, body = await asyncio.to_thread(
        _http_request,
        arguments["url"],
        method="GET",
        headers=arguments.get("headers") or {},
    )
    return {"status_code": status, "headers": headers, "text": truncate_text(body, max_chars)}


async def http_post_json_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
    max_chars = int(arguments.get("max_chars", 12000))
    status, headers, body = await asyncio.to_thread(
        _http_request,
        arguments["url"],
        method="POST",
        headers=arguments.get("headers") or {},
        json_body=arguments.get("json_body") or {},
    )
    return {"status_code": status, "headers": headers, "text": truncate_text(body, max_chars)}


async def web_search_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
    query = str(arguments["query"])
    max_results = int(arguments.get("max_results", 5))
    url = f"https://html.duckduckgo.com/html/?q={quote_plus(query)}"
    try:
        status, headers, body = await asyncio.to_thread(
            _http_request,
            url,
            method="GET",
            headers={
                "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123 Safari/537.36",
            },
        )
    except error.URLError as exc:
        return {"query": query, "results": [], "error": str(exc.reason)}
    results = []
    for match in RESULT_RE.finditer(body):
        href = _normalize_duckduckgo_url(match.group("href"))
        title = _strip_tags(match.group("title"))
        snippet = _strip_tags(match.group("snippet_a") or match.group("snippet_b") or "")
        if not title or not href:
            continue
        results.append({"title": title, "url": href, "snippet": snippet})
        if len(results) >= max_results:
            break
    return {
        "query": query,
        "status_code": status,
        "results": results,
        "raw_preview": truncate_text(body, 1500),
    }



from __future__ import annotations

import pytest

from openclaw_lite.config import default_config
from openclaw_lite.tools.shell import _is_safe_command, run_shell_command


def test_safe_command_allowlist(tmp_path):
    config = default_config(tmp_path)
    assert _is_safe_command("ls -la", config)[0] is True
    allowed, reason = _is_safe_command("rm -rf /", config)
    assert allowed is False
    assert "blocked" in reason


@pytest.mark.asyncio
async def test_run_shell_command_safe(tmp_path):
    config = default_config(tmp_path)
    result = await run_shell_command(config, command="python3 -c 'print(123)'", mode="safe", cwd=tmp_path)
    assert result.returncode == 0
    assert result.stdout.strip() == "123"



from __future__ import annotations

import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from openclaw_lite.schema import ChatMessage
from openclaw_lite.utils import utc_now_iso


@dataclass(slots=True)
class SessionState:
    session_id: str
    provider_name: str
    model: str | None
    mode: str = "safe"
    messages: list[ChatMessage] = field(default_factory=list)
    updated_at: str = field(default_factory=utc_now_iso)
    metadata: dict[str, Any] = field(default_factory=dict)


class SessionStore:
    def __init__(self, base_dir: Path, default_provider: str, default_model: str | None, default_mode: str) -> None:
        self.base_dir = Path(base_dir)
        self.base_dir.mkdir(parents=True, exist_ok=True)
        self.default_provider = default_provider
        self.default_model = default_model
        self.default_mode = default_mode

    def _path(self, session_id: str) -> Path:
        safe = session_id.replace("/", "_")
        return self.base_dir / f"{safe}.json"

    def load(self, session_id: str) -> SessionState:
        path = self._path(session_id)
        if not path.exists():
            return SessionState(
                session_id=session_id,
                provider_name=self.default_provider,
                model=self.default_model,
                mode=self.default_mode,
            )
        raw = json.loads(path.read_text(encoding="utf-8"))
        return SessionState(
            session_id=raw["session_id"],
            provider_name=raw.get("provider_name", self.default_provider),
            model=raw.get("model", self.default_model),
            mode=raw.get("mode", self.default_mode),
            messages=[ChatMessage(**item) for item in raw.get("messages", [])],
            updated_at=raw.get("updated_at", utc_now_iso()),
            metadata=raw.get("metadata", {}),
        )

    def save(self, state: SessionState) -> None:
        state.updated_at = utc_now_iso()
        payload = {
            "session_id": state.session_id,
            "provider_name": state.provider_name,
            "model": state.model,
            "mode": state.mode,
            "updated_at": state.updated_at,
            "metadata": state.metadata,
            "messages": [
                {
                    "role": message.role,
                    "content": message.content,
                    "name": message.name,
                    "tool_call_id": message.tool_call_id,
                }
                for message in state.messages
            ],
        }
        self._path(state.session_id).write_text(json.dumps(payload, ensure_ascii=False, indent=2), encoding="utf-8")

    def trim_history(self, state: SessionState, history_limit: int) -> None:
        if history_limit <= 0:
            return
        state.messages = state.messages[-history_limit:]



from __future__ import annotations

import asyncio
import html
import json
import traceback
import re
from dataclasses import dataclass
from typing import Any

from aiogram import Bot, Dispatcher, F
from aiogram.client.default import DefaultBotProperties
from aiogram.enums import ParseMode
from aiogram.exceptions import TelegramBadRequest
from aiogram.types import BotCommand, CallbackQuery, InlineKeyboardButton, InlineKeyboardMarkup, Message

from ..schemas import TELEGRAM_ACTIVE_USER_ID_METADATA_KEY
from ..tools.telegram_tools import build_tools as build_telegram_tools
from .telegram_media import format_media_prompt_payload, persist_telegram_media
from .telegram_trigger import should_trigger_agent


@dataclass(slots=True)
class ParsedCommand:
    name: str
    args: str = ""
    tokens: tuple[str, ...] = ()
    mention: str | None = None


ADMIN_COMMANDS = {
    "status",
    "safe",
    "mode",
    "providers",
    "models",
    "skills",
    "reload",
    "provider",
    "model",
    "exec",
    "sessions",
    "end",
    "pair",
    "approve",
    "auth",
    "settings",
    "apikey",
    "reset",
}
PUBLIC_COMMANDS = {"start", "help", "pair", "openclaw"}
TAG_RE = re.compile(r"</?[^>]+>")
MARKDOWNISH_CODE_RE = re.compile(r"```([\s\S]*?)```|`([^`\n]+)`", re.S)
BOLD_RE = re.compile(r"\*\*(.+?)\*\*", re.S)
UNDERLINE_RE = re.compile(r"__(.+?)__", re.S)
ITALIC_STAR_RE = re.compile(r"(?<!\*)\*(?!\*)(.+?)(?<!\*)\*(?!\*)", re.S)
ITALIC_UNDERSCORE_RE = re.compile(r"(?<!_)_(?!_)(.+?)(?<!_)_(?!_)", re.S)
STRIKE_RE = re.compile(r"~~(.+?)~~", re.S)
SPOILER_RE = re.compile(r"\|\|(.+?)\|\|", re.S)


def _split_text_chunks(text: str, max_chars: int) -> list[str]:
    if not text:
        return [""]
    return [text[index : index + max_chars] for index in range(0, len(text), max_chars)] or [""]


def _apply_inline_formatting(escaped_text: str) -> str:
    text = BOLD_RE.sub(r"<b>\1</b>", escaped_text)
    text = UNDERLINE_RE.sub(r"<u>\1</u>", text)
    text = STRIKE_RE.sub(r"<s>\1</s>", text)
    text = SPOILER_RE.sub(r"<tg-spoiler>\1</tg-spoiler>", text)
    text = ITALIC_STAR_RE.sub(r"<i>\1</i>", text)
    text = ITALIC_UNDERSCORE_RE.sub(r"<i>\1</i>", text)
    return text


def _split_html_chunks(html_text: str, max_chars: int = 3500) -> list[str]:
    tag_re = re.compile(r"(<[^>]+>)")
    tokens = [token for token in tag_re.split(html_text) if token]
    chunks: list[str] = []
    current = ""
    stack: list[str] = []

    def close_tags() -> str:
        return "".join(f"</{tag}>" for tag in reversed(stack))

    def open_tags() -> str:
        return "".join(f"<{tag}>" for tag in stack)

    for token in tokens:
        if token.startswith("<") and token.endswith(">"):
            if len(current) + len(token) > max_chars and current:
                chunks.append(current + close_tags())
                current = open_tags()
            current += token
            match = re.match(r"</?([a-zA-Z0-9-]+)", token)
            if match:
                tag = match.group(1)
                if token.startswith("</"):
                    if tag in stack[::-1]:
                        idx = len(stack) - 1 - stack[::-1].index(tag)
                        stack = stack[:idx]
                elif not token.endswith("/>"):
                    stack.append(tag)
            continue

        remaining = token
        while remaining:
            space = max_chars - len(current)
            if space <= 0 and current:
                chunks.append(current + close_tags())
                current = open_tags()
                space = max_chars - len(current)
            piece = remaining[:space] if space > 0 else remaining
            if not piece and current:
                chunks.append(current + close_tags())
                current = open_tags()
                continue
            current += piece
            remaining = remaining[len(piece):]
            if remaining:
                chunks.append(current + close_tags())
                current = open_tags()
    if current:
        chunks.append(current + close_tags())
    return chunks or [""]


def render_markdownish_html_chunks(text: str, max_chars: int = 3500) -> list[str]:
    placeholders: dict[str, str] = {}
    counter = 0

    def replace_code(match: re.Match[str]) -> str:
        nonlocal counter
        key = f"@@CODE{counter}@@"
        counter += 1
        fenced = match.group(1)
        inline = match.group(2)
        if fenced is not None:
            placeholders[key] = f"<pre><code>{html.escape(fenced.strip(chr(10)))}</code></pre>"
        else:
            placeholders[key] = f"<code>{html.escape(inline)}</code>"
        return key

    protected = MARKDOWNISH_CODE_RE.sub(replace_code, text)
    escaped = html.escape(protected)
    formatted = _apply_inline_formatting(escaped)
    for key, value in placeholders.items():
        formatted = formatted.replace(html.escape(key), value)
        formatted = formatted.replace(key, value)
    return _split_html_chunks(formatted, max_chars)


def parse_command(text: str | None) -> ParsedCommand | None:
    if not text:
        return None
    stripped = text.strip()
    if not stripped.startswith("/"):
        return None
    head, *tail = stripped.split(maxsplit=1)
    if len(head) <= 1:
        return None
    body = head[1:]
    name, _, mention = body.partition("@")
    args = tail[0].strip() if tail else ""
    tokens = tuple(args.split()) if args else ()
    return ParsedCommand(name=name.lower(), args=args, tokens=tokens, mention=mention or None)


def build_session_id(chat_id: int | str, thread_id: int | None = None) -> str:
    if thread_id is None:
        return f"telegram:{chat_id}:main"
    return f"telegram:{chat_id}:topic:{thread_id}"


def build_help_text() -> str:
    return "\n".join(
        [
            "<b>OpenClaw Lite Telegram</b>",
            "/start, /help",
            "/openclaw &lt;запрос&gt;",
            "/pair — owner: create/list codes | guest: /pair CODE",
            "/status",
            "/safe on|off",
            "/mode safe|yolo|god",
            "/providers",
            "/models [provider]",
            "/provider &lt;name&gt;",
            "/model &lt;name&gt;",
            "/reset [session_id]",
            "/skills",
            "/sessions",
            "/end [session_id]",
            "/approve [user_id] [chat_id]",
            "/auth",
            "/settings",
            "/apikey &lt;provider&gt; &lt;key&gt;",
            "/reload",
            "/exec &lt;command&gt;",
            "/cancel — отменить ввод из /settings",
            "",
            "Slash-подсказки доступны из меню команд Telegram.",
            "Agent mode запускается через reply на бота или /openclaw; пока сессия не закрыта, следующие сообщения того же пользователя продолжают её автоматически.",
        ]
    )


def build_bot_commands() -> list[BotCommand]:
    return [
        BotCommand(command="start", description="старт и краткая помощь"),
        BotCommand(command="help", description="показать список команд"),
        BotCommand(command="openclaw", description="запустить agent mode"),
        BotCommand(command="status", description="статус текущей сессии"),
        BotCommand(command="provider", description="сменить провайдера сессии"),
        BotCommand(command="model", description="сменить модель сессии"),
        BotCommand(command="providers", description="список доступных провайдеров"),
        BotCommand(command="models", description="список моделей"),
        BotCommand(command="settings", description="настройки провайдеров и ключей"),
        BotCommand(command="apikey", description="задать API key для провайдера"),
        BotCommand(command="reset", description="сбросить текущую сессию"),
        BotCommand(command="end", description="закрыть текущую сессию"),
        BotCommand(command="reload", description="перезагрузить runtime"),
        BotCommand(command="cancel", description="отменить ввод из settings"),
        BotCommand(command="skills", description="список загруженных skills"),
        BotCommand(command="pair", description="привязать доступ по pairing code"),
    ]


def _chunk_inline_buttons(buttons: list[InlineKeyboardButton], row_size: int) -> list[list[InlineKeyboardButton]]:
    return [buttons[index : index + row_size] for index in range(0, len(buttons), row_size)]


def build_settings_keyboard(providers: list[dict[str, Any]], *, current_provider: str, current_model: str) -> InlineKeyboardMarkup:
    rows: list[list[InlineKeyboardButton]] = []
    provider_buttons: list[InlineKeyboardButton] = []
    key_buttons: list[InlineKeyboardButton] = []
    model_buttons: list[InlineKeyboardButton] = []
    for item in providers:
        name = str(item.get("name"))
        if not name:
            continue
        mark = "✅ " if name == current_provider else ""
        provider_buttons.append(InlineKeyboardButton(text=f"{mark}{name}", callback_data=f"settings:provider:{name}"))
        key_mark = "✅" if item.get("has_api_key") else "🔑"
        key_buttons.append(InlineKeyboardButton(text=f"{key_mark} key {name}", callback_data=f"settings:key:{name}"))
        model_mark = "✅ " if name == current_provider else ""
        model_buttons.append(
            InlineKeyboardButton(text=f"{model_mark}model {name}", callback_data=f"settings:model:{name}")
        )
    rows.extend(_chunk_inline_buttons(provider_buttons, 3))
    if key_buttons:
        rows.extend(_chunk_inline_buttons(key_buttons, 2))
    if model_buttons:
        rows.extend(_chunk_inline_buttons(model_buttons, 2))
    rows.append([InlineKeyboardButton(text="Refresh", callback_data="settings:refresh")])
    return InlineKeyboardMarkup(inline_keyboard=rows)


def _strip_html_markup(text: str) -> str:
    return TAG_RE.sub("", text)


def _preview(value: Any, limit: int = 700) -> str:
    text = str(value or "")
    if len(text) > limit:
        text = text[:limit] + "…"
    return html.escape(text)


def _status_text(status: dict[str, Any]) -> str:
    lines = [
        "<b>OpenClaw Lite status</b>",
        f"session: <code>{html.escape(str(status.get('session_id')))}</code>",
        f"provider: <code>{html.escape(str(status.get('provider')))}</code>",
        f"model: <code>{html.escape(str(status.get('model')))}</code>",
        f"mode: <code>{html.escape(str(status.get('security_mode')))}</code>",
        f"closed: <code>{html.escape(str(status.get('closed')))}</code>",
        f"history: <code>{html.escape(str(status.get('history_messages')))}</code>",
    ]
    skills = status.get("skills") or []
    if skills:
        skill_names = ", ".join(item.get("name", "?") for item in skills)
        lines.append(f"skills: <code>{html.escape(skill_names)}</code>")
    return "\n".join(lines)


def _providers_text(providers: list[str]) -> str:
    if not providers:
        return "Провайдеры не настроены."
    return "<b>Providers</b>\n" + "\n".join(f"- <code>{html.escape(name)}</code>" for name in providers)


def _skills_text(skills: list[dict[str, Any]]) -> str:
    if not skills:
        return "<b>Skills</b>\n- none"
    lines = ["<b>Skills</b>"]
    for item in skills:
        flag = "on" if item.get("enabled") else "off"
        lines.append(f"- <code>{html.escape(str(item.get('name')))}</code> [{flag}]")
    return "\n".join(lines)


def _models_text(models: Any) -> str:
    if isinstance(models, dict):
        lines = ["<b>Models</b>"]
        for provider, items in models.items():
            lines.append(f"<code>{html.escape(str(provider))}</code>:")
            for model in items:
                lines.append(f"- <code>{html.escape(str(model))}</code>")
        return "\n".join(lines)
    if isinstance(models, list):
        return "<b>Models</b>\n" + "\n".join(f"- <code>{html.escape(str(item))}</code>" for item in models)
    return f"<b>Models</b>\n<code>{html.escape(str(models))}</code>"


def _sessions_text(items: list[dict[str, Any]]) -> str:
    if not items:
        return "<b>Sessions</b>\n- none"
    lines = ["<b>Sessions</b>"]
    for item in items:
        lines.append(
            "- <code>{}</code> | {}/{} | mode={} | closed={}".format(
                html.escape(str(item.get("session_id"))),
                html.escape(str(item.get("provider"))),
                html.escape(str(item.get("model"))),
                html.escape(str(item.get("security_mode"))),
                html.escape(str(item.get("closed"))),
            )
        )
    return "\n".join(lines)


def _auth_text(payload: dict[str, Any]) -> str:
    lines = ["<b>Auth</b>"]
    lines.append(f"owners: <code>{html.escape(str(payload.get('owner_user_ids')))}</code>")
    lines.append(f"owner chats: <code>{html.escape(str(payload.get('owner_chat_ids')))}</code>")
    lines.append(f"allowed users: <code>{html.escape(str(payload.get('allowed_user_ids')))}</code>")
    lines.append(f"allowed chats: <code>{html.escape(str(payload.get('allowed_chat_ids')))}</code>")
    pending = payload.get("pairing_codes") or []
    if pending:
        lines.append("pending codes:")
        for item in pending:
            lines.append(
                f"- <code>{html.escape(str(item.get('code')))}</code> exp=<code>{html.escape(str(item.get('expires_at')))}</code>"
            )
    return "\n".join(lines)


class TelegramProgressReporter:
    def __init__(self, message: Message, bot: Bot):
        self.message = message
        self.bot = bot
        self.status_message: Message | None = None
        self.last_status: str | None = None
        self.last_assistant_text: str | None = None

    async def _send_rendered(self, text: str) -> None:
        for index, chunk in enumerate(render_markdownish_html_chunks(text, 3500)):
            try:
                if index == 0:
                    await self.message.reply(chunk)
                else:
                    await self.bot.send_message(
                        chat_id=self.message.chat.id,
                        text=chunk,
                        reply_to_message_id=self.message.message_id,
                    )
            except TelegramBadRequest:
                plain = _strip_html_markup(chunk)
                if index == 0:
                    await self.message.reply(plain, parse_mode=None)
                else:
                    await self.bot.send_message(
                        chat_id=self.message.chat.id,
                        text=plain,
                        parse_mode=None,
                        reply_to_message_id=self.message.message_id,
                    )

    def _render_status(self, payload: dict[str, Any]) -> str | None:
        event = payload.get("event")
        if event == "agent_iteration_started":
            return (
                "<b>Agent mode</b>\n"
                f"iteration: <code>{payload.get('iteration')}</code>\n"
                f"provider: <code>{html.escape(str(payload.get('provider')))}</code>\n"
                f"model: <code>{html.escape(str(payload.get('model')))}</code>\n"
                "status: thinking"
            )
        if event == "provider_generate_error" and not payload.get("will_retry"):
            return (
                "<b>Agent mode</b>\n"
                f"iteration: <code>{payload.get('iteration')}</code>\n"
                f"provider: <code>{html.escape(str(payload.get('provider')))}</code>\n"
                "status: provider error\n"
                f"error: {_preview(payload.get('error'), 220)}"
            )
        if event == "tool_call_started":
            return (
                "<b>Agent mode</b>\n"
                f"iteration: <code>{payload.get('iteration')}</code>\n"
                f"tool: <code>{html.escape(str(payload.get('tool')))}</code>\n"
                "status: running tool"
            )
        if event == "tool_call_finished":
            result = payload.get("result") or {}
            error = result.get("error") if isinstance(result, dict) else None
            status = "error handled" if error else "tool finished"
            preview = _preview(error or result)
            return (
                "<b>Agent mode</b>\n"
                f"iteration: <code>{payload.get('iteration')}</code>\n"
                f"tool: <code>{html.escape(str(payload.get('tool')))}</code>\n"
                f"status: {status}\n"
                f"preview: {preview}"
            )
        if event == "provider_retry_scheduled":
            delay = payload.get("delay_sec")
            retry_in = f"{delay:.1f}s" if isinstance(delay, (int, float)) else str(delay)
            return (
                "<b>Agent mode</b>\n"
                f"iteration: <code>{payload.get('iteration')}</code>\n"
                f"provider: <code>{html.escape(str(payload.get('provider')))}</code>\n"
                f"model: <code>{html.escape(str(payload.get('model')))}</code>\n"
                f"attempt: <code>{payload.get('attempt')}</code>/<code>{payload.get('max_attempts')}</code>\n"
                f"status: retry in {html.escape(retry_in)}\n"
                f"error: {_preview(payload.get('error'))}"
            )
        if event == "agent_finished":
            status = "session closed" if payload.get("session_closed") else "waiting next message"
            return (
                "<b>Agent mode</b>\n"
                f"iterations: <code>{payload.get('iteration')}</code>\n"
                f"session closed: <code>{html.escape(str(payload.get('session_closed')))}</code>\n"
                f"status: {status}"
            )
        return None

    async def __call__(self, payload: dict[str, Any]) -> None:
        if payload.get("event") == "assistant_response":
            assistant_text = str(payload.get("text") or "").strip()
            if assistant_text and assistant_text != self.last_assistant_text:
                self.last_assistant_text = assistant_text
                await self._send_rendered(assistant_text)
            return
        text = self._render_status(payload)
        if not text or text == self.last_status:
            return
        self.last_status = text
        if self.status_message is None:
            self.status_message = await self.message.reply(text)
            return
        try:
            await self.status_message.edit_text(text)
        except TelegramBadRequest:
            pass

    async def finish(self, payload: dict[str, Any]) -> None:
        if self.status_message is None:
            return
        text = (
            "<b>Agent mode finished</b>\n"
            f"iterations: <code>{html.escape(str(payload.get('iterations')))}</code>\n"
            f"session closed: <code>{html.escape(str(payload.get('session_closed')))}</code>\n"
            f"mode: <code>{html.escape(str(payload.get('security_mode')))}</code>"
        )
        try:
            await self.status_message.edit_text(text)
        except TelegramBadRequest:
            pass


class TelegramBotInterface:
    def __init__(self, runtime) -> None:
        self.runtime = runtime
        telegram = runtime.config.get("telegram", {}) or {}
        token = telegram.get("bot_token") or ""
        if not token:
            raise RuntimeError("telegram.bot_token is not configured")
        parse_mode_name = str(telegram.get("parse_mode", "HTML")).upper()
        parse_mode = ParseMode.HTML if parse_mode_name == "HTML" else ParseMode.MARKDOWN_V2
        self.bot = Bot(token=token, default=DefaultBotProperties(parse_mode=parse_mode))
        self.dp = Dispatcher()
        self.runtime.telegram_bot = self.bot
        self.pending_inputs: dict[tuple[int, int], dict[str, str]] = {}
        self.runtime.add_tool_registrar(lambda registry: registry.register_many(build_telegram_tools()))
        self._register_routes()

    async def _configure_bot_commands(self) -> None:
        try:
            await self.bot.set_my_commands(build_bot_commands())
        except Exception:
            traceback.print_exc()

    def _user_id(self, message: Message) -> int | None:
        return message.from_user.id if message.from_user else None

    def _chat_id(self, message: Message) -> int:
        return int(message.chat.id)

    def _session_id(self, message: Message) -> str:
        return build_session_id(message.chat.id, getattr(message, "message_thread_id", None))

    def _chat_title(self, message: Message) -> str | None:
        return getattr(message.chat, "title", None)

    def _username(self, message: Message) -> str | None:
        user = message.from_user
        if user is None:
            return None
        return user.username or user.full_name

    def _pending_key(self, message: Message, *, user_id: int | None = None) -> tuple[int, int]:
        resolved_user_id = self._user_id(message) if user_id is None else user_id
        return (self._chat_id(message), resolved_user_id or 0)

    def _set_pending_input(self, message: Message, payload: dict[str, str], *, user_id: int | None = None) -> None:
        self.pending_inputs[self._pending_key(message, user_id=user_id)] = payload

    def _peek_pending_input(self, message: Message, *, user_id: int | None = None) -> dict[str, str] | None:
        return self.pending_inputs.get(self._pending_key(message, user_id=user_id))

    def _pop_pending_input(self, message: Message, *, user_id: int | None = None) -> dict[str, str] | None:
        return self.pending_inputs.pop(self._pending_key(message, user_id=user_id), None)

    async def _render_settings(self, message: Message, *, user_id: int | None = None) -> None:
        session = self.runtime.get_session(
            session_id=self._session_id(message),
            chat_id=str(self._chat_id(message)),
            user_id=self._user_id(message) if user_id is None else user_id,
        )
        providers = self.runtime.provider_summary_payload()
        lines = [
            "<b>LiteClaw Settings</b>",
            f"current provider: <code>{html.escape(session.provider)}</code>",
            f"current model: <code>{html.escape(session.model)}</code>",
            "команды: <code>/provider &lt;name&gt;</code> · <code>/model &lt;name&gt;</code> · <code>/apikey &lt;provider&gt; &lt;key&gt;</code> · <code>/reset</code>",
        ]
        for item in providers:
            lines.append(
                f"- <code>{html.escape(str(item['name']))}</code> | enabled={item.get('enabled')} | key={item.get('has_api_key')} | default_model={html.escape(str(item.get('default_model')))}"
            )
        await message.reply("\n".join(lines), reply_markup=build_settings_keyboard(providers, current_provider=session.provider, current_model=session.model))

    async def _handle_pending_input(self, message: Message) -> bool:
        pending = self._peek_pending_input(message)
        if not pending:
            return False
        raw_text = (message.text or message.caption or "").strip()
        if not raw_text:
            await message.reply("Отправь текстом значение для ожидаемой настройки или /cancel")
            return True
        if raw_text == "/cancel":
            self._pop_pending_input(message)
            await message.reply("Отменено.")
            return True
        self._pop_pending_input(message)
        action = pending.get("action")
        provider = pending.get("provider", "")
        session = self.runtime.get_session(session_id=self._session_id(message), chat_id=str(self._chat_id(message)), user_id=self._user_id(message))
        if action == "api_key":
            payload = self.runtime.set_provider_api_key(provider, raw_text)
            await self._reply_json(message, payload)
            await self._render_settings(message)
            return True
        if action == "model":
            self.runtime.set_provider(session, provider)
            self.runtime.set_provider_default_model(provider, raw_text)
            payload = self.runtime.set_model(session, raw_text)
            await self._reply_json(message, payload)
            await self._render_settings(message)
            return True
        return False

    def _is_reply_to_bot(self, message: Message) -> bool:
        reply = message.reply_to_message
        return bool(reply and reply.from_user and reply.from_user.id == self.bot.id)

    def _has_active_agent_session(self, message: Message) -> bool:
        session = self._find_session_by_id(self._session_id(message))
        if session is None:
            return False
        if session.closed:
            return False
        active_user_id = session.metadata.get(TELEGRAM_ACTIVE_USER_ID_METADATA_KEY)
        return active_user_id == self._user_id(message)

    def _has_media(self, message: Message) -> bool:
        return any(
            getattr(message, attr, None)
            for attr in ["photo", "document", "video", "animation", "audio", "voice", "sticker"]
        )

    async def _build_agent_payload(self, message: Message, *, override_text: str | None = None) -> str:
        base_text = (override_text if override_text is not None else (message.text or message.caption or "")).strip()
        if not self._has_media(message):
            return base_text or "[empty message]"
        try:
            media_payload = await persist_telegram_media(self.bot, message, self.runtime.config)
        except Exception as exc:
            traceback.print_exc()
            fallback = base_text or (message.text or message.caption or "").strip()
            return (fallback + f"\n[saved telegram attachments error: {exc}]").strip()
        return format_media_prompt_payload(
            message_text=base_text or media_payload.get("message_text", ""),
            saved_items=media_payload.get("saved_items", []),
        )

    async def _run_agent_request(self, message: Message, *, text: str) -> None:
        reporter = TelegramProgressReporter(message, self.bot)
        result = await self.runtime.handle_user_message(
            session_id=self._session_id(message),
            chat_id=str(message.chat.id),
            user_id=self._user_id(message),
            text=text,
            transport="telegram",
            transport_state={
                "bot": self.bot,
                "message": message,
                "chat_id": message.chat.id,
                "progress_callback": reporter,
            },
        )
        await reporter.finish(result)
        if str(result.get("text") or "").strip() != str(reporter.last_assistant_text or "").strip():
            await self._reply_long(message, result["text"])

    def _is_owner(self, message: Message | None = None, *, user_id: int | None = None, chat_id: int | None = None) -> bool:
        resolved_user_id = user_id if user_id is not None else (self._user_id(message) if message is not None else None)
        resolved_chat_id = chat_id if chat_id is not None else (self._chat_id(message) if message is not None else None)
        return self.runtime.auth_is_owner(user_id=resolved_user_id, chat_id=resolved_chat_id)

    def _is_authorized(self, message: Message | None = None, *, user_id: int | None = None, chat_id: int | None = None) -> bool:
        resolved_user_id = user_id if user_id is not None else (self._user_id(message) if message is not None else None)
        resolved_chat_id = chat_id if chat_id is not None else (self._chat_id(message) if message is not None else None)
        return self.runtime.auth_is_allowed(user_id=resolved_user_id, chat_id=resolved_chat_id)

    def _register_routes(self) -> None:
        @self.dp.message(F.text | F.caption)
        async def on_message(message: Message) -> None:
            await self.handle_message(message)

        @self.dp.callback_query()
        async def on_callback(callback: CallbackQuery) -> None:
            await self.handle_callback(callback)

    async def _reply_long(self, message: Message, text: str, *, html_mode: bool = False) -> None:
        if html_mode:
            plain_text = _strip_html_markup(text)
            chunks = [html.escape(chunk) for chunk in _split_text_chunks(plain_text, 3500)] or [""]
        else:
            chunks = render_markdownish_html_chunks(text, 3500)
        for index, chunk in enumerate(chunks):
            try:
                if index == 0:
                    await message.reply(chunk)
                else:
                    await self.bot.send_message(chat_id=message.chat.id, text=chunk, reply_to_message_id=message.message_id)
            except TelegramBadRequest:
                plain_chunk = _strip_html_markup(chunk)
                if index == 0:
                    await message.reply(plain_chunk, parse_mode=None)
                else:
                    await self.bot.send_message(chat_id=message.chat.id, text=plain_chunk, parse_mode=None, reply_to_message_id=message.message_id)

    async def _reply_json(self, message: Message, payload: Any) -> None:
        text = json.dumps(payload, ensure_ascii=False, indent=2, default=str)
        chunks = [text[i : i + 3200] for i in range(0, len(text), 3200)] or [""]
        for index, chunk in enumerate(chunks):
            rendered = f"<pre>{html.escape(chunk)}</pre>"
            if index == 0:
                await message.reply(rendered)
            else:
                await self.bot.send_message(chat_id=message.chat.id, text=rendered, reply_to_message_id=message.message_id)

    async def _incoming_payload(self, message: Message) -> str:
        return await self._build_agent_payload(message)

    def _find_session_by_id(self, session_id: str):
        for session in self.runtime.session_store.list_sessions():
            if session.session_id == session_id:
                return session
        return None

    async def handle_message(self, message: Message) -> None:
        try:
            text = message.text or message.caption or ""
            parsed = parse_command(text)
            if parsed is not None:
                handled = await self.handle_command(message, parsed)
                if handled:
                    return
            decision = should_trigger_agent(
                text,
                is_reply_to_bot=self._is_reply_to_bot(message),
                has_active_session=self._has_active_agent_session(message),
            )
            if await self._handle_pending_input(message):
                return
            if not decision.accept:
                return
            if not self._is_authorized(message):
                await message.reply("Доступ запрещён. Используй /pair CODE или добавь user/chat в allowlist.")
                return
            await self._run_agent_request(message, text=await self._incoming_payload(message))
        except Exception as exc:
            traceback.print_exc()
            await message.reply(f"Ошибка Telegram handler: <code>{html.escape(str(exc))}</code>")

    async def _handle_pair_command(self, message: Message, command: ParsedCommand) -> bool:
        user_id = self._user_id(message)
        chat_id = self._chat_id(message)
        if not command.args:
            if self._is_owner(message):
                payload = self.runtime.create_pairing_code(created_by=user_id)
                await message.reply(
                    f"Pairing code: <code>{html.escape(str(payload['code']))}</code>\nexpires_at: <code>{html.escape(str(payload['expires_at']))}</code>"
                )
            else:
                await message.reply("Используй: /pair CODE")
            return True

        first = command.tokens[0].lower() if command.tokens else ""
        if self._is_owner(message) and first in {"create", "new"}:
            ttl = int(command.tokens[1]) if len(command.tokens) >= 2 and command.tokens[1].isdigit() else None
            payload = self.runtime.create_pairing_code(created_by=user_id, ttl_seconds=ttl)
            await message.reply(
                f"Pairing code: <code>{html.escape(str(payload['code']))}</code>\nexpires_at: <code>{html.escape(str(payload['expires_at']))}</code>"
            )
            return True

        if self._is_owner(message) and first in {"list", "show"}:
            await self._reply_long(message, _auth_text(self.runtime.auth_summary()), html_mode=True)
            return True

        try:
            payload = self.runtime.pair_telegram_access(
                command.tokens[0],
                user_id=user_id,
                chat_id=chat_id,
                username=self._username(message),
                chat_title=self._chat_title(message),
            )
        except Exception as exc:
            await message.reply(f"Ошибка pairing: <code>{html.escape(str(exc))}</code>")
            return True
        await self._reply_json(message, payload)
        return True

    async def handle_callback(self, callback: CallbackQuery) -> None:
        data = callback.data or ""
        if not data.startswith("settings:"):
            await callback.answer()
            return
        message = callback.message
        if message is None:
            await callback.answer()
            return
        actor_user_id = callback.from_user.id if callback.from_user else None
        if not self._is_owner(user_id=actor_user_id, chat_id=self._chat_id(message)):
            await callback.answer("Эти кнопки доступны только owner.", show_alert=True)
            return
        parts = data.split(":", 2)
        action = parts[1] if len(parts) > 1 else ""
        value = parts[2] if len(parts) > 2 else ""
        session = self.runtime.get_session(
            session_id=self._session_id(message),
            chat_id=str(self._chat_id(message)),
            user_id=actor_user_id,
        )
        if action == "provider":
            payload = self.runtime.set_provider(session, value)
            await callback.answer(f"Provider: {value}")
            await self._reply_json(message, payload)
            await self._render_settings(message, user_id=actor_user_id)
            return
        if action == "key":
            self._set_pending_input(message, {"action": "api_key", "provider": value}, user_id=actor_user_id)
            await callback.answer()
            await message.reply(f"Пришли следующим сообщением API key для <code>{html.escape(value)}</code> или /cancel")
            return
        if action == "model":
            self._set_pending_input(message, {"action": "model", "provider": value}, user_id=actor_user_id)
            await callback.answer()
            await message.reply(f"Пришли следующим сообщением модель для <code>{html.escape(value)}</code> или /cancel")
            return
        if action == "refresh":
            await callback.answer()
            await self._render_settings(message, user_id=actor_user_id)
            return
        await callback.answer()

    async def handle_command(self, message: Message, command: ParsedCommand) -> bool:
        user_id = self._user_id(message)
        chat_id = self._chat_id(message)

        if command.name in {"start", "help"}:
            await message.reply(build_help_text())
            return True

        if command.name == "pair":
            return await self._handle_pair_command(message, command)

        if command.name == "openclaw":
            if not self._is_authorized(message):
                await message.reply("Команда недоступна до авторизации. Используй /pair CODE.")
                return True
            prompt = command.args.strip()
            if not prompt and not self._has_media(message):
                await message.reply("Используй: /openclaw <запрос>, или отправь файл вместе с /openclaw, или reply на сообщение бота.")
                return True
            await self._run_agent_request(message, text=await self._build_agent_payload(message, override_text=prompt))
            return True

        if not self._is_authorized(message):
            await message.reply("Команда недоступна до авторизации. Используй /pair CODE.")
            return True

        if command.name == "auth":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            await self._reply_long(message, _auth_text(self.runtime.auth_summary()), html_mode=True)
            return True

        if command.name == "settings":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            await self._render_settings(message)
            return True

        if command.name == "apikey":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            if len(command.tokens) < 2:
                await message.reply("Используй: /apikey <provider> <key>")
                return True
            provider = command.tokens[0]
            key = command.args.split(None, 1)[1] if len(command.args.split(None, 1)) == 2 else ""
            payload = self.runtime.set_provider_api_key(provider, key)
            await self._reply_json(message, payload)
            return True

        if command.name == "approve":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            target_user = int(command.tokens[0]) if len(command.tokens) >= 1 and command.tokens[0].lstrip("-").isdigit() else user_id
            target_chat = int(command.tokens[1]) if len(command.tokens) >= 2 and command.tokens[1].lstrip("-").isdigit() else chat_id
            payload = self.runtime.approve_telegram_access(
                user_id=target_user,
                chat_id=target_chat,
                approved_by=user_id,
                username=self._username(message),
                chat_title=self._chat_title(message),
            )
            await self._reply_json(message, payload)
            return True

        session = self.runtime.get_session(
            session_id=self._session_id(message),
            chat_id=str(chat_id),
            user_id=user_id,
        )

        if command.name == "status":
            await message.reply(_status_text(self.runtime.get_session_status(session)))
            return True

        if command.name == "safe":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            if not command.tokens:
                await message.reply("Используй: /safe on или /safe off")
                return True
            token = command.tokens[0].lower()
            mode = "safe" if token == "on" else "god" if token == "off" else None
            if mode is None:
                await message.reply("Используй: /safe on или /safe off")
                return True
            result = self.runtime.set_security_mode(session, mode)
            suffix = "\nПолный unrestricted host-terminal access включён." if mode == "god" else ""
            await message.reply(f"Режим: <code>{html.escape(result['security_mode'])}</code>{suffix}")
            return True

        if command.name == "mode":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            if not command.tokens or command.tokens[0].lower() not in {"safe", "yolo", "god"}:
                await message.reply("Используй: /mode safe|yolo|god")
                return True
            result = self.runtime.set_security_mode(session, command.tokens[0].lower())
            await message.reply(f"Режим: <code>{html.escape(result['security_mode'])}</code>")
            return True

        if command.name == "exec":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            if not command.args:
                await message.reply("Используй: /exec <command>")
                return True
            result = await self.runtime.shell_executor.run(command.args, mode=session.security_mode)
            await self._reply_json(message, result.to_dict())
            return True

        if command.name == "providers":
            await message.reply(_providers_text(self.runtime.list_provider_names()))
            return True

        if command.name == "models":
            provider = command.tokens[0] if command.tokens else None
            await self._reply_long(message, _models_text(await self.runtime.list_models(provider)), html_mode=True)
            return True

        if command.name == "provider":
            if not command.tokens:
                await message.reply("Используй: /provider <name>")
                return True
            result = self.runtime.set_provider(session, command.tokens[0])
            await message.reply(
                f"Provider: <code>{html.escape(result['provider'])}</code>\nModel: <code>{html.escape(result['model'])}</code>"
            )
            return True

        if command.name == "model":
            if not command.args:
                await message.reply("Используй: /model <name>")
                return True
            result = self.runtime.set_model(session, command.args)
            await message.reply(f"Model: <code>{html.escape(result['model'])}</code>")
            return True

        if command.name == "reset":
            target_session_id = command.args.strip() if command.args.strip() and self._is_owner(message) else session.session_id
            if target_session_id != session.session_id and self._find_session_by_id(target_session_id) is None:
                await message.reply("Session not found.")
                return True
            if target_session_id == session.session_id:
                self._pop_pending_input(message)
            result = self.runtime.reset_session(target_session_id)
            await message.reply(f"Session reset: <code>{html.escape(result['session_id'])}</code>")
            return True

        if command.name == "skills":
            await message.reply(_skills_text(self.runtime.list_skills_payload()))
            return True

        if command.name == "sessions":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            await self._reply_long(message, _sessions_text(self.runtime.list_sessions_payload()), html_mode=True)
            return True

        if command.name == "end":
            if command.tokens and command.tokens[0].lower() == "reopen":
                target = self._find_session_by_id(command.tokens[1]) if len(command.tokens) >= 2 else session
                if target is None:
                    await message.reply("Session not found.")
                    return True
                result = self.runtime.reopen_session(target)
                await message.reply(f"Session reopened: <code>{html.escape(result['session_id'])}</code>")
                return True
            target = self._find_session_by_id(command.args) if command.args and self._is_owner(message) else session
            if target is None:
                await message.reply("Session not found.")
                return True
            result = self.runtime.end_session(target)
            await message.reply(f"Session ended: <code>{html.escape(result['session_id'])}</code>")
            return True

        if command.name == "reload":
            if not self._is_owner(message):
                await message.reply("Эта команда доступна только owner.")
                return True
            self.runtime.reload()
            await message.reply("Runtime перезагружен.")
            return True

        return False

    async def run_polling(self) -> None:
        await self._configure_bot_commands()
        await self.bot.delete_webhook(drop_pending_updates=True)
        allowed_updates = self.dp.resolve_used_update_types()
        while True:
            try:
                await self.dp.start_polling(self.bot, allowed_updates=allowed_updates)
                break
            except Exception:
                traceback.print_exc()
                await asyncio.sleep(2)

    async def run(self) -> None:
        try:
            await self.run_polling()
        finally:
            await self.bot.session.close()



from __future__ import annotations

from pathlib import Path

from openclaw_lite.config import load_config, save_config


def test_load_config_resolves_env(tmp_path, monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "sk-test")
    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        """
telegram:
  bot_token: ${TELEGRAM_BOT_TOKEN:-abc}
providers:
  default: openai
  items:
    openai:
      type: openai
      api_key: ${OPENAI_API_KEY}
      default_model: gpt-test
""",
        encoding="utf-8",
    )
    config = load_config(config_path)
    assert config.telegram.bot_token == "abc"
    assert config.providers["openai"].api_key == "sk-test"
    assert config.default_model_for("openai") == "gpt-test"


def test_save_and_reload_roundtrip(tmp_path):
    config_path = tmp_path / "config.yaml"
    config = load_config(config_path)
    config.name = "demo"
    save_config(config, config_path)
    reloaded = load_config(config_path)
    assert reloaded.name == "demo"
    assert reloaded.paths.config_file == config_path.resolve()



from __future__ import annotations

from typing import Any

from openclaw_lite.config import save_config, to_dict
from openclaw_lite.schema import ToolContext, ToolSpec
from openclaw_lite.tools.base import fail, ok


def _set_nested(mapping: dict[str, Any], dotted_key: str, value: Any) -> None:
    parts = dotted_key.split(".")
    current = mapping
    for part in parts[:-1]:
        current = current.setdefault(part, {})
    current[parts[-1]] = value


async def runtime_info_tool(context: ToolContext, arguments: dict):
    del arguments
    return ok(
        {
            "session_id": context.session.session_id,
            "provider": context.session.provider_name,
            "model": context.session.model,
            "mode": context.session.mode,
            "paths": {
                "root": str(context.config.paths.root),
                "workspace": str(context.config.paths.workspace),
                "skills": str(context.config.paths.skills),
                "logs": str(context.config.paths.logs),
            },
        }
    )


async def config_get_tool(context: ToolContext, arguments: dict):
    data = to_dict(context.config)
    dotted_key = str(arguments.get("key", "")).strip()
    current: Any = data
    for part in dotted_key.split(".") if dotted_key else []:
        if not isinstance(current, dict) or part not in current:
            return fail(f"config key not found: {dotted_key}")
        current = current[part]
    return ok({"key": dotted_key or "<root>", "value": current})


async def config_set_tool(context: ToolContext, arguments: dict):
    key = str(arguments.get("key", "")).strip()
    value = arguments.get("value")
    if not key:
        return fail("key is required")
    data = to_dict(context.config)
    _set_nested(data, key, value)
    updated = context.runtime.reload_config_from_mapping(data)
    save_config(updated)
    return ok({"key": key, "value": value, "config_file": str(updated.paths.config_file)})


async def set_provider_tool(context: ToolContext, arguments: dict):
    provider_name = str(arguments.get("provider_name", "")).strip()
    if provider_name not in context.runtime.provider_registry.list_names():
        return fail(f"unknown provider: {provider_name}")
    context.session.provider_name = provider_name
    default_model = context.config.default_model_for(provider_name)
    if default_model:
        context.session.model = default_model
    context.runtime.sessions.save(context.session)
    return ok({"provider": context.session.provider_name, "model": context.session.model})


async def set_model_tool(context: ToolContext, arguments: dict):
    model = str(arguments.get("model", "")).strip()
    if not model:
        return fail("model is required")
    context.session.model = model
    context.runtime.sessions.save(context.session)
    return ok({"provider": context.session.provider_name, "model": context.session.model})


async def set_mode_tool(context: ToolContext, arguments: dict):
    mode = str(arguments.get("mode", "")).strip().lower()
    if mode not in {"safe", "yolo", "god"}:
        return fail("mode must be safe, yolo, or god")
    context.session.mode = mode
    context.runtime.sessions.save(context.session)
    return ok({"mode": context.session.mode})


async def reload_runtime_tool(context: ToolContext, arguments: dict):
    del arguments
    summary = await context.runtime.reload_extensions()
    return ok(summary)


async def add_custom_provider_tool(context: ToolContext, arguments: dict):
    name = str(arguments.get("name", "")).strip()
    provider_type = str(arguments.get("type", "openai_compatible")).strip()
    base_url = str(arguments.get("base_url", "")).strip() or None
    api_key = str(arguments.get("api_key", "")).strip()
    default_model = str(arguments.get("default_model", "")).strip() or None
    if not name:
        return fail("name is required")
    context.runtime.add_provider(name=name, provider_type=provider_type, api_key=api_key, base_url=base_url, default_model=default_model)
    save_config(context.config)
    return ok({"provider": name, "type": provider_type, "base_url": base_url, "default_model": default_model})


def build_tools() -> list[ToolSpec]:
    return [
        ToolSpec(
            name="runtime_info",
            description="Return current runtime/session/config information.",
            parameters={"type": "object", "properties": {}},
            handler=runtime_info_tool,
        ),
        ToolSpec(
            name="config_get",
            description="Read a config key by dotted path, or return the full config root.",
            parameters={
                "type": "object",
                "properties": {"key": {"type": "string"}},
            },
            handler=config_get_tool,
        ),
        ToolSpec(
            name="config_set",
            description="Update a config key by dotted path and save config.yaml.",
            parameters={
                "type": "object",
                "properties": {
                    "key": {"type": "string"},
                    "value": {},
                },
                "required": ["key", "value"],
            },
            handler=config_set_tool,
            dangerous=True,
        ),
        ToolSpec(
            name="set_provider",
            description="Switch the active provider for the current session.",
            parameters={
                "type": "object",
                "properties": {"provider_name": {"type": "string"}},
                "required": ["provider_name"],
            },
            handler=set_provider_tool,
        ),
        ToolSpec(
            name="set_model",
            description="Switch the active model for the current session.",
            parameters={
                "type": "object",
                "properties": {"model": {"type": "string"}},
                "required": ["model"],
            },
            handler=set_model_tool,
        ),
        ToolSpec(
            name="set_mode",
            description="Switch shell safety mode: safe, yolo, or god.",
            parameters={
                "type": "object",
                "properties": {"mode": {"type": "string"}},
                "required": ["mode"],
            },
            handler=set_mode_tool,
            dangerous=True,
        ),
        ToolSpec(
            name="reload_runtime",
            description="Reload skills and other runtime extensions without restarting.",
            parameters={"type": "object", "properties": {}},
            handler=reload_runtime_tool,
            dangerous=True,
        ),
        ToolSpec(
            name="add_custom_provider",
            description="Add or replace a custom provider in config and runtime.",
            parameters={
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "type": {"type": "string"},
                    "base_url": {"type": "string"},
                    "api_key": {"type": "string"},
                    "default_model": {"type": "string"},
                },
                "required": ["name", "type"],
            },
            handler=add_custom_provider_tool,
            dangerous=True,
        ),
    ]



from __future__ import annotations

import json
import secrets
from dataclasses import asdict, dataclass, field
from pathlib import Path
from time import time
from typing import Any


@dataclass(slots=True)
class PairingCode:
    code: str
    created_at: float
    expires_at: float
    created_by: int | None = None
    note: str = ""
    max_uses: int = 1
    uses: int = 0

    @property
    def expired(self) -> bool:
        return self.expires_at <= time() or self.uses >= self.max_uses


@dataclass(slots=True)
class AuthGrant:
    user_id: int | None = None
    chat_id: int | None = None
    granted_at: float = field(default_factory=time)
    source: str = "pairing"
    paired_by: int | None = None
    username: str | None = None
    chat_title: str | None = None


class TelegramAccessManager:
    def __init__(self, config: dict[str, Any], state_path: str | Path | None = None):
        self.config = config
        self.state_path = Path(state_path or self._default_state_path(config)).expanduser().resolve()
        self.state_path.parent.mkdir(parents=True, exist_ok=True)
        self.state = self._load_state()

    @staticmethod
    def _default_state_path(config: dict[str, Any]) -> Path:
        data_dir = Path(config.get("data_dir") or Path(config.get("project_root", ".")) / "data")
        return data_dir / "telegram_auth.json"

    def _load_state(self) -> dict[str, Any]:
        if not self.state_path.exists():
            return {"grants": [], "pairing_codes": [], "used_bootstrap_codes": []}
        try:
            data = json.loads(self.state_path.read_text(encoding="utf-8"))
        except json.JSONDecodeError:
            data = {"grants": [], "pairing_codes": [], "used_bootstrap_codes": []}
        data.setdefault("grants", [])
        data.setdefault("pairing_codes", [])
        data.setdefault("used_bootstrap_codes", [])
        return data

    def save(self) -> None:
        self.state_path.write_text(json.dumps(self.state, ensure_ascii=False, indent=2), encoding="utf-8")

    def reload(self) -> dict[str, Any]:
        self.state = self._load_state()
        return self.state

    def _telegram_cfg(self) -> dict[str, Any]:
        return self.config.get("telegram", {}) or {}

    def _pairing_cfg(self) -> dict[str, Any]:
        return self._telegram_cfg().get("pairing", {}) or {}

    def owner_user_ids(self) -> set[int]:
        return {int(item) for item in self._telegram_cfg().get("owner_user_ids", [])}

    def owner_chat_ids(self) -> set[int]:
        return {int(item) for item in self._telegram_cfg().get("owner_chat_ids", [])}

    def allowed_user_ids(self) -> set[int]:
        configured = {int(item) for item in self._telegram_cfg().get("allowed_user_ids", [])}
        granted = {int(item["user_id"]) for item in self.state.get("grants", []) if item.get("user_id") is not None}
        return configured | granted | self.owner_user_ids()

    def allowed_chat_ids(self) -> set[int]:
        configured = {int(item) for item in self._telegram_cfg().get("allowed_chat_ids", [])}
        granted = {int(item["chat_id"]) for item in self.state.get("grants", []) if item.get("chat_id") is not None}
        return configured | granted | self.owner_chat_ids()

    def is_owner(self, user_id: int | None = None, chat_id: int | None = None) -> bool:
        if user_id is not None and int(user_id) in self.owner_user_ids():
            return True
        if chat_id is not None and int(chat_id) in self.owner_chat_ids():
            return True
        return False

    def is_allowed(self, user_id: int | None = None, chat_id: int | None = None) -> bool:
        if self.is_owner(user_id=user_id, chat_id=chat_id):
            return True
        if user_id is not None and int(user_id) in self.allowed_user_ids():
            return True
        if chat_id is not None and int(chat_id) in self.allowed_chat_ids():
            return True
        return False

    def is_approved(self, user_id: int | None = None, chat_id: int | None = None) -> bool:
        return self.is_allowed(user_id=user_id, chat_id=chat_id)

    def pairing_enabled(self) -> bool:
        return bool(self._pairing_cfg().get("enabled", True))

    def pairing_code_ttl_sec(self) -> int:
        pairing_cfg = self._pairing_cfg()
        return int(pairing_cfg.get("code_ttl_sec", pairing_cfg.get("ttl_seconds", 900)))

    def bootstrap_code(self) -> str:
        return str(self._pairing_cfg().get("bootstrap_code", "") or "").strip().upper()

    def bootstrap_one_time(self) -> bool:
        return bool(self._pairing_cfg().get("bootstrap_one_time", True))

    def _pairing_codes(self) -> list[PairingCode]:
        return [PairingCode(**item) for item in self.state.get("pairing_codes", [])]

    def _set_pairing_codes(self, codes: list[PairingCode]) -> None:
        self.state["pairing_codes"] = [asdict(item) for item in codes]

    def cleanup_expired_codes(self) -> int:
        current = self._pairing_codes()
        kept = [item for item in current if not item.expired]
        removed = len(current) - len(kept)
        if removed:
            self._set_pairing_codes(kept)
            self.save()
        return removed

    def create_pairing_code(
        self,
        *,
        created_by: int | None = None,
        expires_in: int | None = None,
        note: str = "",
        max_uses: int = 1,
    ) -> PairingCode:
        if not self.pairing_enabled():
            raise RuntimeError("pairing is disabled")
        self.cleanup_expired_codes()
        ttl = int(expires_in if expires_in is not None else self.pairing_code_ttl_sec())
        now = time()
        code = PairingCode(
            code=secrets.token_hex(3).upper(),
            created_at=now,
            expires_at=now + max(ttl, 1),
            created_by=created_by,
            note=note,
            max_uses=max(max_uses, 1),
        )
        codes = self._pairing_codes()
        codes.append(code)
        self._set_pairing_codes(codes)
        self.save()
        return code

    def generate_pairing_code(
        self,
        *,
        created_by: int | None = None,
        note: str | None = None,
        ttl_seconds: int | None = None,
    ) -> dict[str, Any]:
        return asdict(self.create_pairing_code(created_by=created_by, expires_in=ttl_seconds, note=note or ""))

    def _find_code(self, code: str) -> tuple[int, PairingCode] | tuple[None, None]:
        lookup = code.strip().upper()
        codes = self._pairing_codes()
        for index, item in enumerate(codes):
            if item.code.upper() == lookup:
                return index, item
        return None, None

    def list_pairing_codes(self) -> list[PairingCode]:
        self.cleanup_expired_codes()
        return self._pairing_codes()

    def pair_with_code(
        self,
        code: str,
        *,
        user_id: int | None = None,
        chat_id: int | None = None,
        username: str | None = None,
        chat_title: str | None = None,
    ) -> AuthGrant:
        self.cleanup_expired_codes()
        index, pairing_code = self._find_code(code)
        if pairing_code is None or index is None:
            raise ValueError("invalid pairing code")
        if pairing_code.expired:
            raise ValueError("pairing code expired")
        grant = AuthGrant(
            user_id=int(user_id) if user_id is not None else None,
            chat_id=int(chat_id) if chat_id is not None else None,
            granted_at=time(),
            source="pairing_code",
            paired_by=pairing_code.created_by,
            username=username,
            chat_title=chat_title,
        )
        self.add_grant(grant)
        codes = self._pairing_codes()
        codes[index].uses += 1
        if codes[index].expired:
            del codes[index]
        self._set_pairing_codes(codes)
        self.save()
        return grant

    def claim_pairing_code(self, code: str, *, user_id: int | None, chat_id: int | None) -> dict[str, Any]:
        normalized = code.strip().upper()
        if self.bootstrap_code() and normalized == self.bootstrap_code():
            if self.bootstrap_one_time() and normalized in set(self.state.get("used_bootstrap_codes", [])):
                raise ValueError("bootstrap pairing code already used")
            payload = self.approve_identity(user_id=user_id, chat_id=chat_id, via="bootstrap", code=normalized)
            if self.bootstrap_one_time():
                used = set(self.state.get("used_bootstrap_codes", []))
                used.add(normalized)
                self.state["used_bootstrap_codes"] = sorted(used)
                self.save()
            payload.setdefault("approved", True)
            payload.setdefault("via", payload.get("source", "bootstrap"))
            payload.setdefault("code", normalized)
            return payload
        grant = asdict(self.pair_with_code(code, user_id=user_id, chat_id=chat_id))
        grant.setdefault("via", grant.get("source"))
        grant["approved"] = True
        return grant

    def add_grant(self, grant: AuthGrant) -> AuthGrant:
        grants = self.state.get("grants", [])
        normalized = asdict(grant)
        normalized.setdefault("via", normalized.get("source"))
        duplicate = any(
            item.get("user_id") == normalized.get("user_id") and item.get("chat_id") == normalized.get("chat_id")
            for item in grants
        )
        if not duplicate:
            grants.append(normalized)
            self.state["grants"] = grants
            self.save()
        return grant

    def approve_identity(
        self,
        *,
        user_id: int | None = None,
        chat_id: int | None = None,
        via: str = "manual",
        paired_by: int | None = None,
        username: str | None = None,
        chat_title: str | None = None,
        code: str | None = None,
    ) -> dict[str, Any]:
        grant = AuthGrant(
            user_id=int(user_id) if user_id is not None else None,
            chat_id=int(chat_id) if chat_id is not None else None,
            source=via,
            paired_by=paired_by,
            username=username,
            chat_title=chat_title,
        )
        self.add_grant(grant)
        payload = asdict(grant)
        if code is not None:
            payload["code"] = code
        payload["approved"] = True
        return payload

    def revoke_user(self, user_id: int) -> int:
        before = len(self.state.get("grants", []))
        self.state["grants"] = [item for item in self.state.get("grants", []) if item.get("user_id") != int(user_id)]
        removed = before - len(self.state["grants"])
        if removed:
            self.save()
        return removed

    def revoke_chat(self, chat_id: int) -> int:
        before = len(self.state.get("grants", []))
        self.state["grants"] = [item for item in self.state.get("grants", []) if item.get("chat_id") != int(chat_id)]
        removed = before - len(self.state["grants"])
        if removed:
            self.save()
        return removed

    def revoke(self, *, user_id: int | None = None, chat_id: int | None = None) -> dict[str, Any]:
        removed = 0
        if user_id is not None:
            removed += self.revoke_user(int(user_id))
        if chat_id is not None:
            removed += self.revoke_chat(int(chat_id))
        return {"revoked": removed > 0, "removed": removed, "user_id": user_id, "chat_id": chat_id}

    def export_summary(self) -> dict[str, Any]:
        self.cleanup_expired_codes()
        grants = []
        for item in self.state.get("grants", []):
            payload = dict(item)
            payload.setdefault("via", payload.get("source", "manual"))
            grants.append(payload)
        pairing_codes = [asdict(item) for item in self._pairing_codes()]
        return {
            "owner_user_ids": sorted(self.owner_user_ids()),
            "owner_chat_ids": sorted(self.owner_chat_ids()),
            "allowed_user_ids": sorted(self.allowed_user_ids()),
            "allowed_chat_ids": sorted(self.allowed_chat_ids()),
            "grants": grants,
            "pending_codes": pairing_codes,
            "pairing_codes": pairing_codes,
            "state_path": str(self.state_path),
            "bootstrap_code_configured": bool(self.bootstrap_code()),
        }

    def summary(self) -> dict[str, Any]:
        return self.export_summary()

    def export_state(self) -> dict[str, Any]:
        return self.export_summary()



from __future__ import annotations

import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Any


def utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def json_dumps(data: Any) -> str:
    return json.dumps(data, ensure_ascii=False, indent=2, sort_keys=True)


def shorten(text: str, limit: int) -> str:
    if len(text) <= limit:
        return text
    return text[: limit - 3] + "..."


def ensure_text(value: Any) -> str:
    if isinstance(value, str):
        return value
    return json_dumps(value)


def ensure_parent(path: Path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)



from __future__ import annotations

from typing import Any


async def shell_run_handler(context, arguments: dict[str, Any]) -> dict[str, Any]:
    mode = arguments.get("mode") or context.session.security_mode
    result = await context.runtime.shell_executor.run(
        arguments["command"],
        mode=mode,
        cwd=arguments.get("cwd"),
        timeout_seconds=arguments.get("timeout_seconds"),
    )
    return result.to_dict()



from __future__ import annotations

from typing import Any

from openclaw_lite.schema import ToolResult


def object_schema(properties: dict[str, Any], required: list[str] | None = None) -> dict[str, Any]:
    return {
        "type": "object",
        "properties": properties,
        "required": required or [],
        "additionalProperties": False,
    }


def ok(content: Any = None, *, metadata: dict[str, Any] | None = None) -> ToolResult:
    return ToolResult(ok=True, content=content, metadata=metadata or {})


def fail(error: str, *, content: Any = None, metadata: dict[str, Any] | None = None) -> ToolResult:
    return ToolResult(ok=False, content=content, error=str(error), metadata=metadata or {})



from __future__ import annotations

import html
import threading
import webbrowser
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from urllib.parse import parse_qs

from ..config import ConfigManager


_PROVIDER_PRESETS: dict[str, dict[str, str]] = {
    "openai": {
        "label": "OpenAI",
        "api_key_label": "OpenAI API Key",
        "type": "openai",
        "default_model": "gpt-4.1-mini",
        "description": "ChatGPT-style API with the built-in OpenAI provider.",
    },
    "gemini": {
        "label": "Gemini",
        "api_key_label": "Gemini API Key",
        "type": "gemini",
        "default_model": "gemini-2.5-flash",
        "description": "Google Gemini models through the native Gemini provider.",
    },
    "claude": {
        "label": "Claude",
        "api_key_label": "Claude API Key",
        "type": "claude",
        "default_model": "claude-3-7-sonnet-latest",
        "description": "Anthropic Claude models through the built-in Claude provider.",
    },
}


def _parse_ids(raw: str) -> list[int]:
    result = []
    for chunk in raw.replace("\n", ",").split(","):
        item = chunk.strip()
        if item and item.lstrip("-").isdigit():
            result.append(int(item))
    return result


def _config_telegram(config: dict) -> dict:
    return config.get("telegram", {}) or {}


def _config_provider_items(config: dict) -> dict[str, dict]:
    providers = config.get("providers", {}) or {}
    items = providers.get("items", {}) or {}
    return items if isinstance(items, dict) else {}


def _provider_entry(config: dict, name: str) -> dict:
    entry = _config_provider_items(config).get(name, {}) or {}
    return entry if isinstance(entry, dict) else {}


def _default_provider(config: dict, allowed: set[str] | None = None) -> str:
    provider = str((config.get("providers", {}) or {}).get("default", "openai") or "openai")
    allowed_names = allowed or set(_PROVIDER_PRESETS)
    if provider in allowed_names:
        return provider
    return "openai" if "openai" in allowed_names else next(iter(allowed_names))


def _provider_label(name: str) -> str:
    preset = _PROVIDER_PRESETS.get(name, {})
    return str(preset.get("label", name.title()))


def _provider_model(config: dict, name: str) -> str:
    preset = _PROVIDER_PRESETS.get(name, {})
    return str(_provider_entry(config, name).get("default_model") or preset.get("default_model", ""))


def _provider_api_key(config: dict, name: str) -> str:
    return str(_provider_entry(config, name).get("api_key", ""))


def _provider_base_url(config: dict, name: str) -> str:
    return str(_provider_entry(config, name).get("base_url", ""))


def _messages_html(*, notice: str = "", errors: list[str] | None = None) -> str:
    chunks: list[str] = []
    if notice:
        chunks.append(f"<div class='flash flash-ok'>{html.escape(notice)}</div>")
    if errors:
        items = "".join(f"<li>{html.escape(item)}</li>" for item in errors)
        chunks.append(f"<div class='flash flash-error'><ul>{items}</ul></div>")
    return "".join(chunks)


def _page_shell(*, title: str, subtitle: str, body: str, notice: str = "", errors: list[str] | None = None) -> str:
    return f"""<!doctype html>
<html>
<head>
  <meta charset='utf-8'>
  <meta name='viewport' content='width=device-width, initial-scale=1'>
  <title>{html.escape(title)}</title>
  <style>
    :root {{
      color-scheme: dark;
      --bg: #09111f;
      --panel: rgba(12, 19, 35, 0.88);
      --panel-2: rgba(16, 24, 43, 0.94);
      --line: rgba(148, 163, 184, 0.18);
      --text: #e5eefb;
      --muted: #9fb0ce;
      --accent: #74c0fc;
      --accent-2: #8b5cf6;
      --ok: #123c2a;
      --ok-line: #2e8b57;
      --error: #43181b;
      --error-line: #dc2626;
    }}
    * {{ box-sizing: border-box; }}
    body {{
      margin: 0;
      min-height: 100vh;
      font-family: Inter, system-ui, -apple-system, BlinkMacSystemFont, 'Segoe UI', sans-serif;
      background:
        radial-gradient(circle at top right, rgba(116, 192, 252, 0.15), transparent 32%),
        radial-gradient(circle at top left, rgba(139, 92, 246, 0.15), transparent 26%),
        linear-gradient(180deg, #08101d 0%, #0b1324 100%);
      color: var(--text);
    }}
    .page {{ max-width: 1120px; margin: 0 auto; padding: 32px 18px 48px; }}
    .hero {{
      padding: 28px;
      border: 1px solid var(--line);
      border-radius: 28px;
      background: linear-gradient(180deg, rgba(17, 25, 45, 0.94), rgba(11, 18, 32, 0.96));
      box-shadow: 0 20px 48px rgba(0, 0, 0, 0.24);
    }}
    .eyebrow {{
      display: inline-flex;
      align-items: center;
      gap: 8px;
      padding: 6px 12px;
      border-radius: 999px;
      background: rgba(116, 192, 252, 0.12);
      color: #cfe8ff;
      font-size: 12px;
      letter-spacing: 0.08em;
      text-transform: uppercase;
    }}
    h1 {{ margin: 14px 0 10px; font-size: 36px; line-height: 1.06; }}
    .subtitle {{ margin: 0; color: var(--muted); max-width: 760px; line-height: 1.6; }}
    .content {{ margin-top: 22px; display: grid; gap: 18px; }}
    .panel {{
      border: 1px solid var(--line);
      border-radius: 22px;
      padding: 22px;
      background: var(--panel);
      backdrop-filter: blur(10px);
    }}
    .panel h2 {{ margin: 0 0 6px; font-size: 20px; }}
    .panel > p {{ margin: 0 0 18px; color: var(--muted); }}
    .grid {{ display: grid; gap: 16px; grid-template-columns: repeat(2, minmax(0, 1fr)); }}
    .providers {{ display: grid; gap: 16px; grid-template-columns: repeat(auto-fit, minmax(240px, 1fr)); }}
    .provider-card {{
      border: 1px solid rgba(148, 163, 184, 0.16);
      border-radius: 18px;
      padding: 16px;
      background: var(--panel-2);
    }}
    .provider-choice {{ display: flex; align-items: center; gap: 10px; margin-bottom: 12px; font-weight: 700; }}
    .provider-choice input {{ width: 18px; height: 18px; margin: 0; accent-color: var(--accent); }}
    .provider-card p {{ margin: 0 0 14px; color: var(--muted); min-height: 40px; }}
    label {{ display: block; margin-top: 12px; font-weight: 600; color: #dce8fb; }}
    input, textarea, select {{
      width: 100%;
      margin-top: 6px;
      border-radius: 14px;
      border: 1px solid rgba(148, 163, 184, 0.18);
      padding: 12px 14px;
      background: rgba(4, 10, 22, 0.72);
      color: var(--text);
      font: inherit;
    }}
    textarea {{ min-height: 104px; resize: vertical; }}
    .hint {{ color: var(--muted); font-size: 12px; line-height: 1.45; margin-top: 6px; }}
    .flash {{ border-radius: 16px; padding: 14px 16px; margin-bottom: 18px; }}
    .flash-ok {{ background: var(--ok); border: 1px solid var(--ok-line); color: #ccf0da; }}
    .flash-error {{ background: var(--error); border: 1px solid var(--error-line); color: #fecaca; }}
    .flash-error ul {{ margin: 0; padding-left: 18px; }}
    .actions {{ display: flex; flex-wrap: wrap; gap: 12px; margin-top: 22px; }}
    button {{
      appearance: none;
      border: 0;
      border-radius: 999px;
      padding: 12px 18px;
      font: inherit;
      font-weight: 700;
      cursor: pointer;
    }}
    .btn-primary {{
      color: #04121f;
      background: linear-gradient(135deg, #74c0fc, #8b5cf6 120%);
    }}
    .btn-secondary {{
      color: var(--text);
      background: rgba(148, 163, 184, 0.12);
      border: 1px solid rgba(148, 163, 184, 0.18);
    }}
    .footer-note {{ color: var(--muted); font-size: 13px; margin-top: 14px; }}
    @media (max-width: 860px) {{
      h1 {{ font-size: 30px; }}
      .grid {{ grid-template-columns: 1fr; }}
    }}
  </style>
</head>
<body>
  <div class='page'>
    <section class='hero'>
      <div class='eyebrow'>LiteClaw setup</div>
      <h1>{html.escape(title)}</h1>
      <p class='subtitle'>{html.escape(subtitle)}</p>
    </section>
    <section class='content'>
      <div class='panel'>
        {_messages_html(notice=notice, errors=errors)}
        {body}
      </div>
    </section>
  </div>
</body>
</html>"""


def render_settings_page(config: dict, *, notice: str = "", errors: list[str] | None = None) -> str:
    telegram = _config_telegram(config)
    default_provider = _default_provider(config, allowed={"openai", "gemini", "claude", "openrouter"})
    openai = _provider_entry(config, "openai")
    gemini = _provider_entry(config, "gemini")
    claude = _provider_entry(config, "claude")
    openrouter = _provider_entry(config, "openrouter")
    provider_options = "".join(
        f"<option value='{html.escape(name)}'{' selected' if name == default_provider else ''}>{html.escape(_provider_label(name))}</option>"
        for name in ["openai", "gemini", "claude", "openrouter"]
    )
    body = f"""
      <h2>Settings</h2>
      <p>Edit Telegram access, choose the default provider, and manage provider API keys.</p>
      <form method='post' action='/save'>
        <div class='grid'>
          <div>
            <label>Telegram Bot Token
              <input type='password' name='telegram_bot_token' value='{html.escape(str(telegram.get("bot_token", "")))}' autocomplete='off'>
            </label>
            <label>Owner User IDs
              <input type='text' name='telegram_owner_user_ids' value='{html.escape(",".join(str(x) for x in telegram.get("owner_user_ids", [])))}'>
              <div class='hint'>Comma-separated Telegram user ids.</div>
            </label>
            <label>Allowed User IDs
              <input type='text' name='telegram_allowed_user_ids' value='{html.escape(",".join(str(x) for x in telegram.get("allowed_user_ids", [])))}'>
            </label>
            <label>Allowed Chat IDs
              <input type='text' name='telegram_allowed_chat_ids' value='{html.escape(",".join(str(x) for x in telegram.get("allowed_chat_ids", [])))}'>
            </label>
          </div>
          <div>
            <label>Default Provider
              <select name='default_provider'>{provider_options}</select>
            </label>
            <label>OpenRouter API Key
              <input type='password' name='openrouter_api_key' value='{html.escape(str(openrouter.get("api_key", "")))}' autocomplete='off'>
            </label>
            <label>OpenRouter Base URL
              <input type='text' name='openrouter_base_url' value='{html.escape(str(openrouter.get("base_url", "")))}'>
            </label>
            <label>OpenRouter Default Model
              <input type='text' name='openrouter_default_model' value='{html.escape(str(openrouter.get("default_model", "")))}'>
            </label>
          </div>
        </div>
        <div class='providers'>
          <div class='provider-card'>
            <div class='provider-choice'>OpenAI</div>
            <p>Built-in OpenAI provider with the standard OpenAI API.</p>
            <label>OpenAI API Key
              <input type='password' name='openai_api_key' value='{html.escape(str(openai.get("api_key", "")))}' autocomplete='off'>
            </label>
            <label>OpenAI Default Model
              <input type='text' name='openai_default_model' value='{html.escape(str(openai.get("default_model", "")))}'>
            </label>
          </div>
          <div class='provider-card'>
            <div class='provider-choice'>Gemini</div>
            <p>Built-in Gemini provider for Google models.</p>
            <label>Gemini API Key
              <input type='password' name='gemini_api_key' value='{html.escape(str(gemini.get("api_key", "")))}' autocomplete='off'>
            </label>
            <label>Gemini Default Model
              <input type='text' name='gemini_default_model' value='{html.escape(str(gemini.get("default_model", "")))}'>
            </label>
          </div>
          <div class='provider-card'>
            <div class='provider-choice'>Claude</div>
            <p>Built-in Claude provider for Anthropic models.</p>
            <label>Claude API Key
              <input type='password' name='claude_api_key' value='{html.escape(str(claude.get("api_key", "")))}' autocomplete='off'>
            </label>
            <label>Claude Default Model
              <input type='text' name='claude_default_model' value='{html.escape(str(claude.get("default_model", "")))}'>
            </label>
          </div>
        </div>
        <div class='actions'>
          <button type='submit' class='btn-primary'>Save settings</button>
        </div>
      </form>
    """
    return _page_shell(
        title="OpenClaw Lite Settings",
        subtitle="Update tokens, access lists, and provider settings without editing YAML by hand.",
        body=body,
        notice=notice,
        errors=errors,
    )


def render_onboarding_page(config: dict, *, notice: str = "", errors: list[str] | None = None) -> str:
    telegram = _config_telegram(config)
    default_provider = _default_provider(config)
    cards = []
    for name, preset in _PROVIDER_PRESETS.items():
        checked = " checked" if name == default_provider else ""
        cards.append(
            f"""
            <article class='provider-card'>
              <label class='provider-choice'>
                <input type='radio' name='default_provider' value='{html.escape(name)}'{checked}>
                <span>{html.escape(preset['label'])}</span>
              </label>
              <p>{html.escape(preset['description'])}</p>
              <label>{html.escape(preset['label'])} API Key
                <input type='password' name='{html.escape(name)}_api_key' value='{html.escape(_provider_api_key(config, name))}' autocomplete='off'>
              </label>
              <label>{html.escape(preset['label'])} Default Model
                <input type='text' name='{html.escape(name)}_default_model' value='{html.escape(_provider_model(config, name))}'>
              </label>
            </article>
            """
        )
    body = f"""
      <h2>Quickstart onboard</h2>
      <p>Paste the Telegram bot token, choose your main model provider, add its API key, and LiteClaw can launch right after saving.</p>
      <form method='post' action='/save'>
        <div class='grid'>
          <div>
            <label>Telegram Bot Token
              <input type='password' name='telegram_bot_token' value='{html.escape(str(telegram.get("bot_token", "")))}' autocomplete='off'>
            </label>
            <label>Owner User IDs
              <input type='text' name='telegram_owner_user_ids' value='{html.escape(",".join(str(x) for x in telegram.get("owner_user_ids", [])))}'>
              <div class='hint'>Optional, but recommended so admin commands are immediately available to you.</div>
            </label>
          </div>
          <div>
            <label>Allowed User IDs
              <input type='text' name='telegram_allowed_user_ids' value='{html.escape(",".join(str(x) for x in telegram.get("allowed_user_ids", [])))}'>
            </label>
            <label>Allowed Chat IDs
              <input type='text' name='telegram_allowed_chat_ids' value='{html.escape(",".join(str(x) for x in telegram.get("allowed_chat_ids", [])))}'>
            </label>
            <p class='footer-note'>Choose one provider card below as default. LiteClaw will save the provider key and default model into `config.yaml`.</p>
          </div>
        </div>
        <div class='providers'>
          {''.join(cards)}
        </div>
        <div class='actions'>
          <button type='submit' name='submit_action' value='launch' class='btn-primary'>Save and launch LiteClaw</button>
          <button type='submit' name='submit_action' value='save' class='btn-secondary'>Save draft only</button>
        </div>
        <p class='footer-note'>After launch, keep this terminal open. The web panel shuts down automatically once the runtime starts.</p>
      </form>
    """
    return _page_shell(
        title="LiteClaw onboard",
        subtitle="First-run setup for Telegram, provider selection, API keys, and one-click launch.",
        body=body,
        notice=notice,
        errors=errors,
    )


def _apply_provider(items: dict[str, dict], name: str, *, provider_type: str, api_key: str, default_model: str, base_url: str | None = None) -> None:
    entry = dict(items.get(name, {}) or {})
    entry["type"] = provider_type
    entry["enabled"] = True
    entry["api_key"] = api_key
    entry["default_model"] = default_model
    if base_url is not None:
        entry["base_url"] = base_url
    items[name] = entry


def _apply_shared_form_update(data: dict, form: dict[str, str]) -> dict:
    telegram = data.setdefault("telegram", {})
    telegram["bot_token"] = form.get("telegram_bot_token", "")
    telegram["owner_user_ids"] = _parse_ids(form.get("telegram_owner_user_ids", ""))
    telegram["allowed_user_ids"] = _parse_ids(form.get("telegram_allowed_user_ids", ""))
    telegram["allowed_chat_ids"] = _parse_ids(form.get("telegram_allowed_chat_ids", ""))

    providers = data.setdefault("providers", {})
    items = providers.setdefault("items", {})
    selected_provider = form.get("default_provider") or providers.get("default") or "openai"
    providers["default"] = selected_provider

    _apply_provider(
        items,
        "openai",
        provider_type="openai",
        api_key=form.get("openai_api_key", str((items.get("openai") or {}).get("api_key", ""))),
        default_model=form.get("openai_default_model", "") or str((items.get("openai") or {}).get("default_model", _PROVIDER_PRESETS["openai"]["default_model"])),
    )
    _apply_provider(
        items,
        "gemini",
        provider_type="gemini",
        api_key=form.get("gemini_api_key", str((items.get("gemini") or {}).get("api_key", ""))),
        default_model=form.get("gemini_default_model", "") or str((items.get("gemini") or {}).get("default_model", _PROVIDER_PRESETS["gemini"]["default_model"])),
    )
    _apply_provider(
        items,
        "claude",
        provider_type="claude",
        api_key=form.get("claude_api_key", str((items.get("claude") or {}).get("api_key", ""))),
        default_model=form.get("claude_default_model", "") or str((items.get("claude") or {}).get("default_model", _PROVIDER_PRESETS["claude"]["default_model"])),
    )
    _apply_provider(
        items,
        "openrouter",
        provider_type="openai_compatible",
        api_key=form.get("openrouter_api_key", str((items.get("openrouter") or {}).get("api_key", ""))),
        default_model=form.get("openrouter_default_model", "") or str((items.get("openrouter") or {}).get("default_model", "")),
        base_url=form.get("openrouter_base_url", str((items.get("openrouter") or {}).get("base_url", ""))),
    )
    return data


def update_config_from_form(manager: ConfigManager, form: dict[str, str]) -> dict:
    manager.backup("web-ui")
    _apply_shared_form_update(manager.data, form)
    manager.save()
    return manager.data


def update_onboarding_from_form(manager: ConfigManager, form: dict[str, str]) -> dict:
    manager.backup("onboard")
    _apply_shared_form_update(manager.data, form)
    manager.save()
    return manager.data


def validate_onboarding_form(form: dict[str, str]) -> list[str]:
    errors: list[str] = []
    if not form.get("telegram_bot_token", "").strip():
        errors.append("Telegram Bot Token is required.")
    selected_provider = (form.get("default_provider") or "openai").strip()
    if selected_provider not in _PROVIDER_PRESETS:
        errors.append("Choose a supported default provider.")
        return errors
    api_key = form.get(f"{selected_provider}_api_key", "").strip()
    if not api_key:
        errors.append(f"{_PROVIDER_PRESETS[selected_provider]['api_key_label']} is required.")
    return errors


def _browser_url(host: str, port: int) -> str:
    display_host = "127.0.0.1" if host in {"0.0.0.0", "::"} else host
    return f"http://{display_host}:{port}"


def _try_open_browser(url: str) -> None:
    try:
        webbrowser.open(url, new=1)
    except Exception:
        return


def _run_ui_server(
    config_path: str,
    *,
    host: str,
    port: int,
    page_renderer,
    save_handler,
    launch_handler=None,
    banner: str,
    open_browser: bool,
) -> bool:
    manager = ConfigManager(config_path)
    state = {"launch_requested": False}

    class Handler(BaseHTTPRequestHandler):
        def _send_html(self, body: str, status: int = 200) -> None:
            data = body.encode("utf-8")
            self.send_response(status)
            self.send_header("Content-Type", "text/html; charset=utf-8")
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data)

        def do_GET(self) -> None:
            if self.path not in {"/", "/index.html"}:
                self._send_html("<h1>Not Found</h1>", 404)
                return
            manager.reload()
            self._send_html(page_renderer(manager.data))

        def do_POST(self) -> None:
            if self.path != "/save":
                self._send_html("<h1>Not Found</h1>", 404)
                return
            length = int(self.headers.get("Content-Length", "0"))
            raw = self.rfile.read(length).decode("utf-8")
            parsed = {key: values[-1] for key, values in parse_qs(raw, keep_blank_values=True).items()}
            if launch_handler is not None:
                notice, errors, should_launch = launch_handler(manager, parsed)
            else:
                notice, errors, should_launch = save_handler(manager, parsed)
            self._send_html(page_renderer(manager.data, notice=notice, errors=errors))
            if should_launch:
                state["launch_requested"] = True
                threading.Thread(target=server.shutdown, daemon=True).start()

        def log_message(self, format, *args):
            return

    def settings_post(manager: ConfigManager, parsed: dict[str, str]) -> tuple[str, list[str] | None, bool]:
        save_handler(manager, parsed)
        return ("Saved. Use /reload in Telegram or restart runtime.", None, False)

    if launch_handler is None:
        launch_handler = settings_post

    server = ThreadingHTTPServer((host, port), Handler)
    browser_url = _browser_url(host, port)
    try:
        print(f"{banner}: {browser_url}")
        if open_browser:
            threading.Thread(target=_try_open_browser, args=(browser_url,), daemon=True).start()
        server.serve_forever()
    finally:
        server.server_close()
    return bool(state["launch_requested"])


def run_web_ui(config_path: str, host: str = "127.0.0.1", port: int = 8787) -> None:
    _run_ui_server(
        config_path,
        host=host,
        port=port,
        page_renderer=render_settings_page,
        save_handler=update_config_from_form,
        banner="OpenClaw Lite web UI",
        open_browser=False,
    )


def run_onboarding_ui(config_path: str, host: str = "127.0.0.1", port: int = 8787, *, open_browser: bool = True) -> bool:
    def handle_post(manager: ConfigManager, parsed: dict[str, str]) -> tuple[str, list[str] | None, bool]:
        action = parsed.get("submit_action", "save")
        if action == "launch":
            errors = validate_onboarding_form(parsed)
            if errors:
                manager.reload()
                return ("", errors, False)
            update_onboarding_from_form(manager, parsed)
            return ("Saved. Launching LiteClaw now...", None, True)
        update_onboarding_from_form(manager, parsed)
        return ("Saved. Draft updated — launch whenever you're ready.", None, False)

    return _run_ui_server(
        config_path,
        host=host,
        port=port,
        page_renderer=render_onboarding_page,
        save_handler=update_onboarding_from_form,
        launch_handler=handle_post,
        banner="LiteClaw onboard UI",
        open_browser=open_browser,
    )
